import os
import re
import sqlite3
import subprocess
from lxml import etree
from Bio import SeqIO
from statistics import median

from app.dataformats import mzidtsv as constants
from tests.integration import basetests


class MzidWithDB(basetests.MzidTSVBaseTest):
    def setUp(self):
        super().setUp()
        self.dbfile = os.path.join(self.fixdir, self.dbfn)
        self.workdb = os.path.join(self.workdir, self.dbfn)
        self.copy_db_to_workdir(self.dbfn, self.workdb)


class TestPSMTable(MzidWithDB):
    command = 'psmtable'
    infilename = 'target.tsv'
    dbfn = 'quant_lookup.sqlite'
    """DB and PSM table contain:
    - ENSEMBL proteins
    - a Uniprot swiss protein
    - A self-annotated protein
    - A non-annotated (only peptide) proteins
    """

    def test_build_full_psmtable(self):
        minpif = '0.4'
        fastafn = os.path.join(self.basefixdir, 'ens99_small.fasta')
        options = ['--dbfile', self.workdb, '--spectracol', '1', '--addmiscleav',
                '--addbioset', '--genes', '--proteingroup', '--ms1quant', '--isobaric',
                '--fasta', fastafn, '--min-precursor-purity', minpif]
        self.run_command(options)
        self.check_db_fasta(fastafn)
        self.check_addspec_miscleav_bioset()
        self.check_pg()
        self.check_quanttsv(minpif)
        self.check_addgenes()

    def test_ionmobility(self):
        self.infilename = 'few_spec_timstof.tsv'
        self.dbfn = 'spectra_lookup_timstof.sqlite'
        self.setUp()
        options = ['--dbfile', self.workdb, '--spectracol', '1', '--addmiscleav', '--addbioset']
        self.run_command(options)
        sql = ('SELECT pr.rownr, bs.set_name, sp.retention_time, '
               'iit.ion_injection_time, im.ion_mobility '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'JOIN mzml AS sp USING(spectra_id) '
               'LEFT OUTER JOIN ioninjtime AS iit USING(spectra_id) '
               'LEFT OUTER JOIN ionmob AS im USING(spectra_id) '
               'JOIN mzmlfiles USING(mzmlfile_id) '
               'JOIN biosets AS bs USING(set_id) '
               'ORDER BY pr.rownr')
        fields = ['Biological set', 'Retention time(min)',
                  'Ion injection time(ms)', 'Ion mobility(Vs/cm2)']
        expected_values = self.process_dbvalues_both(self.workdb, sql,
                                                     [1, 2, 3, 4], fields)
        self.check_results_sql(fields, expected_values)
        for val, exp in zip(self.get_values(['missed_cleavage']), self.get_values(['Peptide'], self.infile[0])):
            exp = re.sub('[0-9\+\.]', '', exp[0][0])[:-1]
            self.assertEqual(int(val[0][0]), exp.count('K') + exp.count('R') - exp.count('KP') - exp.count('RP'))

    def test_no_isoquant_in_some_rows(self):
        '''Test when not all rows of PSMs have isoquant data, e.g. when some are
        CID.'''
        # delete isoquant for first scan in PSM table
        db = sqlite3.connect(self.workdb)
        spec_id_to_rm = '1_controllerType=0 controllerNumber=1 scan=10029'
        db.execute(f'DELETE FROM isobaric_quant WHERE spectra_id="{spec_id_to_rm}"')
        db.execute(f'DELETE FROM precursor_ion_fraction WHERE spectra_id="{spec_id_to_rm}"')
        db.commit()
        self.test_build_full_psmtable()

    def check_addspec_miscleav_bioset(self):
        sql = ('SELECT pr.rownr, bs.set_name, sp.retention_time, '
               'iit.ion_injection_time, im.ion_mobility, pif.pif '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'JOIN mzml AS sp USING(spectra_id) '
               'JOIN ioninjtime AS iit USING(spectra_id) '
               'LEFT OUTER JOIN ionmob AS im USING(spectra_id) '
               'LEFT OUTER JOIN precursor_ion_fraction AS pif USING(spectra_id) '
               'JOIN mzmlfiles USING(mzmlfile_id) '
               'JOIN biosets AS bs USING(set_id) '
               'ORDER BY pr.rownr')
        fields = ['Biological set', 'Retention time(min)',
                  'Ion injection time(ms)', 'Ion mobility(Vs/cm2)',
                  'Precursor ion fraction']
        expected_values = self.process_dbvalues_both(self.workdb, sql,
                                                     [1, 2, 3, 4, 5], fields)
        self.check_results_sql(fields, expected_values)
        # FIXME
        for val, exp in zip(self.get_values(['missed_cleavage']), self.get_values(['Peptide'], self.infile[0])):
            exp = re.sub('[0-9\+\.]', '', exp[0][0])[:-1]
            self.assertEqual(int(val[0][0]), exp.count('K') + exp.count('R') - exp.count('KP') - exp.count('RP'))

    def check_quanttsv(self, minpif):
        sql = ('SELECT pr.rownr, ic.channel_name, '
                'CASE WHEN pif.pif > {} THEN iq.intensity ELSE "NA" END AS intensity '
                'FROM psmrows AS pr JOIN psms USING(psm_id) '
                'LEFT OUTER JOIN precursor_ion_fraction AS pif USING(spectra_id) '
                'JOIN isobaric_quant AS iq USING(spectra_id) '
                'JOIN isobaric_channels AS ic USING(channel_id) '.format(float(minpif)))
        fields = ['tmt10plex_{}'.format(ch) for ch in ['126', '127N', '127C',
                                                       '128N', '128C', '129N',
                                                       '129C', '130N', '130C',
                                                       '131']]
        expected_values = self.rowify(self.get_values_from_db(self.workdb, sql))
        self.check_results_sql(fields, expected_values)
        sql = ('SELECT pr.rownr, pq.intensity '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'LEFT OUTER JOIN ms1_align USING(spectra_id) '
               'LEFT OUTER JOIN ms1_quant AS pq USING(feature_id)')
        expected_values = self.rowify(self.get_values_from_db(self.workdb, sql))
        self.check_results_sql(['MS1 area'], expected_values)

    def check_db_fasta(self, fasta, exp_proteins=None, desc=True):
        if exp_proteins is None:
            exp_proteins = {}
            for rec in SeqIO.parse(fasta, 'fasta'):
                rd = rec.description
                gene = 'NA'
                if 'gene_symbol:' in rd:
                    six = rd.index('gene_symbol:') + 12
                    gix = rd.index('gene:') + 5 
                    gene = rd[gix: rd.index(' ', gix)]
                    desc = rd[rd.index('description:') + 12:]
                elif 'GN=' in rd:
                    six = rd.index('GN=') + 3
                    rsp = rd.split()[1:]
                    nextterm = [i for i,x in enumerate(rsp) if '=' in x][0]
                    desc = ' '.join(rsp[:nextterm])
                elif 'msstitch_fake_gene' in rd:
                    # special case fake fasta record for non-standard gene 
                    six = False
                    desc = rd.split()[-1]
                elif 'msstitch_fake_onlypeptide' in rd:
                    # special fake fasta record for unannotated peptide
                    six, desc = False, 'NA'
                exp_proteins[rec.id] = {
                        'seq': rec.seq,
                        'gene': gene,
                        'desc': desc,
                        'symbol': rd[six: rd.index(' ', six)] if six else 'NA',
                        }
        self.check_db_base(exp_proteins)
        sql = ('SELECT ps.protein_acc, ps.sequence, g.gene_acc, aid.assoc_id, '
                'pd.description '
                'FROM proteins AS p '
                'LEFT OUTER JOIN ensg_proteins USING(pacc_id) '
                'LEFT OUTER JOIN genes AS g USING(gene_id) '
                'LEFT OUTER JOIN genename_proteins AS gnp USING(pacc_id) '
                'LEFT OUTER JOIN associated_ids AS aid USING(gn_id) '
               ' JOIN protein_seq AS ps ON ps.protein_acc=p.protein_acc '
               ' JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id ')
        if not desc:
            sql = ('SELECT ps.protein_acc, ps.sequence '
                   'FROM protein_seq AS ps '
                   'JOIN prot_desc AS pd USING(protein_acc)')
        for prot, seq, gene, aid, desc in self.get_values_from_db(self.workdb,
                                                             sql):
            self.assertEqual(exp_proteins[prot]['seq'], seq)
            if desc:
                self.assertEqual(exp_proteins[prot]['gene'], gene or 'NA')
                self.assertEqual(exp_proteins[prot]['symbol'], aid or 'NA')
                self.assertEqual(exp_proteins[prot]['desc'], desc)

    def check_db_base(self, expected_proteins=None):
        expected_psms = self.get_expected_psms()
        if expected_proteins is None:
            expected_proteins = {x for y in expected_psms.values()
                                 for x in y['proteins']}
        protsql = 'SELECT * FROM proteins'
        for protein in self.get_values_from_db(self.workdb, protsql):
            self.assertIn(protein[1], expected_proteins)
        psmsql = ('SELECT ps.sequence, p.score, pr.rownr '
                  'FROM psmrows AS pr JOIN psms AS p USING(psm_id) '
                  'JOIN peptide_sequences AS ps USING(pep_id)')
        for psm in self.get_values_from_db(self.workdb, psmsql):
            expected_psm = (expected_psms[psm[2]]['seq'],
                            expected_psms[psm[2]]['score'])
            self.assertEqual((psm[0], psm[1]), expected_psm)
        ppsql = ('SELECT pp.protein_acc, pr.rownr FROM psmrows AS pr '
                 'JOIN protein_psm AS pp USING(psm_id)')
        for protpsm in self.get_values_from_db(self.workdb, ppsql):
            self.assertIn(protpsm[0], expected_psms[protpsm[1]]['proteins'])

    def get_expected_psms(self):
        header = self.get_tsvheader(self.infile[0])
        prot_ix = header.index('Protein')
        seq_ix = header.index('Peptide')
        score_ix = header.index('MSGFScore')
        psms = {}
        for row, line in enumerate(self.get_all_lines(self.infile[0])):
            line = line.strip('\n').split('\t')
            psms[row] = {'proteins': [x.split('(pre')[0] for x in
                                      line[prot_ix].split(';')],
                         'seq': line[seq_ix],
                         'score': line[score_ix],
                         }
        return psms

    def parse_proteingroups(self, fn):
        with open(fn) as fp:
            header = next(fp).strip().split('\t')
            master_ix = header.index(constants.HEADER_MASTER_PROT)
            pgcontent_ix = header.index(constants.HEADER_PG_CONTENT)
            pgamount_ix = header.index(constants.HEADER_PG_AMOUNT_PROTEIN_HITS)
            for line in fp:
                line = line.strip().split('\t')
                yield {'master': line[master_ix],
                       'content': line[pgcontent_ix],
                       'amount': line[pgamount_ix],
                       }

    def do_asserting(self, result, expected, unrolled=False):
        for res, exp in zip(result, expected):
            self.assertEqual(set(res['master'].split(';')),
                             set(exp['master'].split(';')))
            self.assertEqual(res['amount'], exp['amount'])
            rescontent = res['content'].split(';')
            expcontent = exp['content'].split(';')
            self.assertEqual(set(rescontent), set(expcontent))

    def check_pglup(self, sql, keyfun, valfun):
        result = {keyfun(x): valfun(x) for x in
                  self.get_values_from_db(self.workdb, sql)}
        exp_file = os.path.join(self.fixdir, 'target_psms.sqlite')
        expected = {keyfun(x): valfun(x) for x in
                    self.get_values_from_db(exp_file, sql)}
        for key, value in result.items():
            self.assertIn(key, expected.keys())
            self.assertEqual(value, expected[key])

    def check_pg(self):
        sql = 'SELECT * FROM protein_coverage'
        self.check_pglup(sql, lambda x: x[0], lambda x: x[1])
        sql = """SELECT ppg.psm_id, p.protein_acc FROM psm_protein_groups
               AS ppg JOIN protein_group_master AS pgm USING(master_id)
               JOIN proteins AS p ON pgm.pacc_id=p.pacc_id"""
        self.check_pglup(sql, lambda x: x[0], lambda x: x[1])
        sql = ('SELECT p.protein_acc, pgc.protein_acc, pgc.peptide_count, '
               'pgc.psm_count, pgc.protein_score '
               'FROM protein_group_content AS pgc '
               'JOIN protein_group_master AS pgm USING(master_id) '
               'JOIN proteins AS p USING(pacc_id) '
               'ORDER BY p.pacc_id, pgc.protein_acc')
        self.check_pglup(sql, lambda x: x[0], lambda x: x[1:])
        sql = ('SELECT * FROM protein_group_master')
        self.check_pglup(sql, lambda x: x[1], lambda x: 1)
        # Check the output TSV
        result = self.parse_proteingroups(self.resultfn)
        expected = self.parse_proteingroups(
            os.path.join(self.fixdir, 'target_pg.tsv'))
        self.do_asserting(result, expected)

    def check_addgenes(self):
        for line in self.get_values(['Gene ID', 'Gene Name', 'Description',
                                     'Protein']):
            genes = line[0][1].split(';')
            assoc_ids = line[1][1].split(';')
            descriptions = ['{}]'.format(x).replace(']]', ']')
                            for x in line[2][1].split('];')]
            proteins = [x.split('(')[0] for x in line[3][1].split(';')]
            sql = ('SELECT p.protein_acc, g.gene_acc, a.assoc_id, '
                   'd.description FROM proteins AS p '
                   'JOIN ensg_proteins USING(pacc_id) '
                   'JOIN genename_proteins USING(pacc_id) '
                   'JOIN genes AS g USING(gene_id) '
                   'JOIN associated_ids AS a USING(gn_id) '
                   ' JOIN prot_desc AS d ON d.pacc_id=p.pacc_id '
                   'WHERE p.protein_acc IN ({})')
            dbvals = self.get_values_from_db(self.workdb, sql.format(
                ','.join(['"{}"'.format(x) for x in proteins])))
            exp_g, exp_assoc, exp_desc = set(), set(), set()
            for prot, gene, aid, desc in dbvals:
                exp_g.add(gene)
                exp_assoc.add(aid)
                exp_desc.add(desc)
            for exp_set, result in zip([exp_g, exp_assoc, exp_desc],
                                       [genes, assoc_ids, descriptions]):
                self.assertEqual(0, len(exp_set.difference(result)))


class TestPercoTSV(basetests.MzidTSVBaseTest):
    command = 'perco2psm'
    suffix = '_fdr.tsv'
    infilename = 'few_spectra.tsv'

    def test_conffilt(self):
        threshold = 0.3
        mzidfn = os.path.join(self.fixdir, 'few_spectra.mzid')
        percofn = os.path.join(self.fixdir, 'perco.xml')
        options = ['--mzid', mzidfn, '--perco', percofn, '--filtpep', 
            str(threshold), '--filtpsm', str(threshold)]
        self.run_command(options)
        checkfields = ['percolator svm-score', 'PSM q-value', 'peptide q-value',
            'PSM PEP', 'peptide PEP', 'TD']
        with open(os.path.join(self.fixdir, 'few_spectra.tsv_fdr.tsv')) as fp:
            header = next(fp).strip().split('\t')
            expected = [line.strip().split('\t') for line in fp]
        expected = [{field: line[i] for i, field in enumerate(header)} for line in expected]
        expected = [line for line in expected if float(line['PSM q-value']) < threshold and
            float(line['peptide q-value']) < threshold]
        # check how many PSMs made the cut:
        self.assertEqual(len(expected),  len([x for x in self.get_values(checkfields)]))
        for res, exp in zip(self.get_values(checkfields), expected):
            for i, field in enumerate(checkfields):
                self.assertEqual(field, res[i][0])
                self.assertEqual(exp[field], res[i][1])

    def test_add_fdr(self):
        mzidfn = os.path.join(self.fixdir, 'few_spectra.mzid')
        percofn = os.path.join(self.fixdir, 'perco.xml')
        options = ['--mzid', mzidfn, '--perco', percofn]
        self.run_command(options)
        checkfields = ['percolator svm-score', 'PSM q-value', 'peptide q-value',
            'PSM PEP', 'peptide PEP', 'TD']
        with open(os.path.join(self.fixdir, 'few_spectra.tsv_fdr.tsv')) as fp:
            header = next(fp).strip().split('\t')
            expected = [line.strip().split('\t') for line in fp]
        expected = [{field: line[i] for i, field in enumerate(header)} for line in expected]
        for res, exp in zip(self.get_values(checkfields), expected):
            print(res, exp)
            for i, field in enumerate(checkfields):
                self.assertEqual(field, res[i][0])
                self.assertEqual(exp[field], res[i][1])

    def test_use_qvality_tsv(self):
        '''Get FDR data from recalculation in qvality. Qvality input is prepared
        as a TSV with header, for target and decoy (so qvality -d) . In this case 
        I have simply taken the percolator svm/qval/pep scores from perco.xml and 
        added 0.1 to each PSM (not peptide) q-value using bash:
        cat <(echo $'Score\tPEP\tq-value') \
                <(grep -A3 '<psm p' perco.xml| grep -v psm| grep -v '\-\-' | \
                sed -E 's/\s*<[\/]*[a-z_]+>//g' | paste - - - ) | \
                awk -v FS='\t' -v OFS='\t' '{print $1, $3, $2+0.1}' > qvality_psms.txt 
        '''
        mzidfn = os.path.join(self.fixdir, 'few_spectra.mzid')
        percofn = os.path.join(self.fixdir, 'perco.xml')
        qval_psms = os.path.join(self.fixdir, 'qvality_psms.txt')
        qval_peps = os.path.join(self.fixdir, 'qvality_peps.txt')
        options = ['--mzid', mzidfn, '--perco', percofn, '--qvalitypsms', qval_psms, '--qvalitypeps',
                qval_peps]
        self.run_command(options)
        checkfields = ['percolator svm-score', 'PSM q-value', 'peptide q-value',
            'PSM PEP', 'peptide PEP', 'TD']
        with open(os.path.join(self.fixdir, 'few_spectra.tsv_fdr.tsv')) as fp:
            header = next(fp).strip().split('\t')
            expected = [line.strip().split('\t') for line in fp]
        expected = [{field: line[i] for i, field in enumerate(header)} for line in expected]
        for res, exp in zip(self.get_values(checkfields), expected):
            for i, field in enumerate(checkfields):
                if field == 'PSM q-value':
                    self.assertAlmostEqual(float(exp[field]) + 0.1, float(res[i][1]))
                else:
                    self.assertEqual(exp[field], res[i][1])
                self.assertEqual(field, res[i][0])


class TestPercoTSVTIMS(basetests.MzidTSVBaseTest):
    command = 'perco2psm'
    suffix = '_fdr.tsv'
    infilename = 'few_spec_timstof.tsv'
    dbfn = 'spectra_lookup_timstof.sqlite'

    def test_add_fdr_timstof(self):
        mzidfn = os.path.join(self.fixdir, 'few_spec_timstof.mzid')
        percofn = os.path.join(self.fixdir, 'perco_timstof.xml')
        options = ['--mzid', mzidfn, '--perco', percofn]
        self.run_command(options)
        with open(os.path.join(self.fixdir, 'few_spec_timstof.tsv_fdr.tsv')) as fp:
            header = next(fp).strip().split('\t')
            expected = [line.strip().split('\t') for line in fp]
        expected = [{field: line[i] for i, field in enumerate(header)} for line in expected]
        for res, exp in zip(self.get_values(header), expected):
            for i, field in enumerate(header):
                self.assertEqual(field, res[i][0])
                self.assertEqual(exp[field], res[i][1])


class TestConcatTSV(basetests.MzidTSVBaseTest):
    command = 'concat'
    suffix = '_concat.tsv'
    infilename = 'few_spectra.tsv'

    def test_mergetsv(self):
        self.infile = [self.infile, self.infile]
        self.run_command()
        resultlines = self.get_all_lines(self.resultfn)
        for expectfn in self.infile:
            for line in self.get_all_lines(expectfn):
                self.assertEqual(line, next(resultlines))


class TestSplitTSV(basetests.MzidTSVBaseTest):
    infilename = 'target_pg.tsv'
    command = 'split'
    suffix = '_split.tsv'

    def setUp(self):
        super().setUp()
        self.resultfn = None
        self.expectlines = [x for x in self.get_all_lines(self.infile)]

    def test_splitcol_bioset(self):
        self.run_command(['--splitcol', 'bioset'])
        for resultfn in [os.path.join(self.workdir, '{}.tsv'.format(x)) for x in ['Set1', 'Set2']]:
            for line in self.get_all_lines(resultfn):
                self.assertIn(line, self.expectlines)

    def test_invalid_splitcol(self):
        options = ['--splitcol', 'hello']
        result = self.run_command(options, return_error=True)
        if result.returncode != 0:
            self.assertEqual(result.stdout.strip(), 
                    'ERROR: --splitcol must be an integer or "TD", or "bioset"')
        else:
            self.fail('This test should error')

    def test_splitcol(self):
        setnames = ['Set1', 'Set2']
        options = ['--splitcol', '30']
        self.run_command(options)
        resultfiles = [os.path.join(self.workdir, '{}.tsv'.format(setname))
                       for setname in setnames]
        for resultfn in resultfiles:
            for line in self.get_all_lines(resultfn):
                self.assertIn(line, self.expectlines)


class TestSeqMatchFastaDB(basetests.MzidTSVBaseTest):
    command = 'seqmatch'
    infilename = 'few_spectra.tsv'
    suffix = '_seqmatch.txt'

    def check_peps_in_out(self, options, seqs, matching=True):
        matchcol = 'TESTMATCH'
        self.run_command([*options, '--dbfile', 'seqs.db', '--matchcolname', matchcol])
        pepfound = False
        seqs_nrmap = {x: ix for ix,x in enumerate(seqs)}
        with open(self.resultfn) as fp:
            header = next(fp).strip().split('\t')
            out_linecount = 0
            for line in fp:
                out_linecount += 1
                psm = {k: v for k,v in zip(header, line.strip().split('\t'))}
                seq = re.sub('[^A-Z]', '', psm['Peptide'])
                if seq in seqs_nrmap and matching:
                    self.assertEqual(psm[matchcol], f'acc_{seqs_nrmap[seq]}')
                else:
                    self.assertEqual(psm[matchcol], 'No match')
        with open(self.infile[0]) as fp:
            header = next(fp).strip().split('\t')
            in_linecount = 0
            for line in fp:
                in_linecount += 1
        self.assertEqual(in_linecount, out_linecount)

    def test_noflags(self):
        seqs = ['AIASWNR']
        basetests.create_db(seqs, mapaccession=True)
        options = ['--dbfile', 'seqs.db']
        self.check_peps_in_out(options, seqs)

    def test_ntermwildcards(self):
        seqs = ['XXAIASWNR']
        seqs_to_filter = ['AIASWNR']
        basetests.create_db(seqs, reverse=True, mapaccession=True)

        # Find peptide
        max_falloff = 2
        options = ['--insourcefrag', str(max_falloff)]
        self.check_peps_in_out(options, seqs_to_filter)

# FIXME
#        # Now failing, too short falloff
#        max_falloff = 1
#        options = ['--insourcefrag', str(max_falloff)]
#        self.check_peps_in_out(options, seqs_to_filter)

    def test_deamidate(self):
        # deamidation: N -> D
        seqs = ['NIENLR']
        seqs_to_filter = ['DIENLR']
        basetests.create_db(seqs, mapaccession=True)
        options = ['--deamidate']
        self.check_peps_in_out(options, seqs_to_filter)

    def test_fullprotein(self):
        '''Testing if PSMs with sequence match (except Isoleucine) are removed'''
        seqs = ['XXXXXXXXXXFMPNNLERYYYYYYY']
        seqs_to_filter = ['FMPNNIER']
        basetests.create_db(seqs, fullprotein=True, minlen=6, mapaccession=True)
        # Find peptide
        options = ['--fullprotein', '--minlen', '6']
        self.check_peps_in_out(options, seqs_to_filter)

    def test_fullprotein_minlenwrong(self):
        '''Do not filter since even with matching 6-aa peptide, the 7th of the PSM does not match'''
        seqs = ['XXXXXXXXXXALASWNYYYYYYY']
        seqs_to_filter = ['AIASWNR']
        basetests.create_db(seqs, fullprotein=True, minlen=6, mapaccession=True)
        # Find peptide
        options = ['--fullprotein', '--minlen', '6']
        self.check_peps_in_out(options, seqs_to_filter, matching=False)



class TestRemoveDupPSM(basetests.MzidTSVBaseTest):
    command = 'deduppsms'
    infilename = 'few_spectra_duplicate.tsv'
    suffix = '_dedup.txt'

    def get_line_count(self, fn):
        with open(fn) as fp:
            return sum(1 for _line in fp)

    def test_nodup_noremove(self):
        options = []
        self.infile = [os.path.join(self.fixdir, 'few_spectra.tsv')]
        self.run_command(options)
        self.assertEqual(self.get_line_count(self.infile[0]), self.get_line_count(self.resultfn))

    def test_removedup(self):
        '''There are two duplicated PSMs in the input file, which is derived from
        the few_spectra.tsv file'''
        options = []
        self.run_command(options)
        self.assertEqual(self.get_line_count(self.infile[0]) - 2, self.get_line_count(self.resultfn))
        checkfile = os.path.join(self.fixdir, 'few_spectra.tsv')
        self.assertEqual(os.path.getsize(checkfile), os.path.getsize(self.resultfn))

    def test_removedup_seqcol(self):
        '''Some PSMs are already duplicate by Scan Number since they have 
        two equal scoring solutions from search engine by sequence'''
        self.infile = [os.path.join(self.fixdir, 'few_spectra.tsv')]
        options = ['--peptidecolpattern', 'ScanNum']
        self.run_command(options)
        self.assertEqual(self.get_line_count(self.infile[0]) - 3, self.get_line_count(self.resultfn))


class TestSeqFilt(basetests.MzidTSVBaseTest):
    command = 'seqfilt'
    infilename = 'few_spectra.tsv'
    suffix = '_filtseq.txt'

    def test_noflags(self):
        seqs = ['AIASWNR']
        basetests.create_db(seqs)
        options = ['--dbfile', 'seqs.db']
        self.run_command(options)
        self.check_peps_in_out(seqs, matching=True)

    def check_peps_in_out(self, seqs, matching):
        '''Check if peptides in seqs are contained in new PSM tables.
        matching is 0/1 - True: peptides passed match, should NOT be found anymore
                          False: peptides do not match, should be found in DB
        '''
        pepfound = False
        with open(self.resultfn) as fp:
            header = next(fp).strip().split('\t')
            filtpsms = {}
            for line in fp:
                psm = {k: v for k,v in zip(header, line.strip().split('\t'))}
                seq = re.sub('[^A-Z]', '', psm['Peptide'])
                scan = psm['ScanNum']
                filtpsms[scan] = seq
                if seq in seqs:
                    pepfound = True
                    break
        if matching:
            self.assertFalse(pepfound)
        else:
            self.assertTrue(pepfound)
        with open(self.infile[0]) as fp:
            allpsms = {}
            header = next(fp).strip().split('\t')
            for line in fp:
                psm = {k: v for k,v in zip(header, line.strip().split('\t'))}
                seq = re.sub('[^A-Z]', '', psm['Peptide'])
                scan = psm['ScanNum']
                allpsms[scan] = seq
                if seq in seqs:
                    pepfound = True
                    break
                else:
                    self.assertEqual(seq, filtpsms[scan])
        self.assertTrue(pepfound)

    def test_ntermwildcards(self):
        seqs = ['XXAIASWNR']
        seqs_to_filter = ['AIASWNR']
        basetests.create_db(seqs, reverse=True)

        # Find peptide
        max_falloff = 2
        options = ['--dbfile', 'seqs.db', '--insourcefrag', str(max_falloff)]
        self.run_command(options)
        self.check_peps_in_out(seqs_to_filter, matching=True)

        # Now failing, too short falloff
        max_falloff = 1
        options = ['--dbfile', 'seqs.db', '--insourcefrag', str(max_falloff)]
        self.run_command(options)
        self.check_peps_in_out(seqs_to_filter, matching=False)

    def test_deamidate(self):
        # deamidation: N -> D
        seqs = ['NIENLR']
        seqs_to_filter = ['DIENLR']
        basetests.create_db(seqs)
        options = ['--dbfile', 'seqs.db', '--deamidate']
        self.run_command(options)
        self.check_peps_in_out(seqs_to_filter, matching=True)

    def test_fullprotein(self):
        '''Testing if PSMs with sequence match (except Isoleucine) are removed'''
        seqs = ['XXXXXXXXXXFMPNNLERYYYYYYY']
        seqs_to_filter = ['FMPNNIER']
        basetests.create_db(seqs, fullprotein=True, minlen=6)
        # Find peptide
        options = ['--dbfile', 'seqs.db', '--fullprotein', '--minlen', '6']
        self.run_command(options)
        self.check_peps_in_out(seqs_to_filter, matching=True)

    def test_fullprotein_minlenwrong(self):
        '''Do not filter since even with matching 6-aa peptide, the 7th of the PSM does not match'''
        seqs = ['XXXXXXXXXXALASWNYYYYYYY']
        seqs_to_filter = ['AIASWNR']
        basetests.create_db(seqs, fullprotein=True, minlen=6)
        # Find peptide
        options = ['--dbfile', 'seqs.db', '--fullprotein', '--minlen', '6']
        self.run_command(options)
        self.check_peps_in_out(seqs_to_filter, matching=False)


class TestConffiltTSV(basetests.MzidTSVBaseTest):
    command = 'conffilt'
    infilename = 'few_spectra.tsv'
    suffix = '_filtconf.txt'

    def test_confidence_filter_lower(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'lower', confcol=14)

    def test_confidence_filter_lower_confpattern(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'lower', confpat='EValue')

    def test_confidence_filter_higher(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'higher', confcol=14)

    def run_conffilt(self, conflvl, better, confcol=False, confpat=False):
        options = ['--confidence-better', better,
                   '--confidence-lvl', str(conflvl)]
        if confcol is not False:
            options.extend(['--confidence-col', str(confcol)])
        elif confpat:
            options.extend(['--confcolpattern', confpat])
        self.run_command(options)
        asserter = {'lower': self.assertLess,
                    'higher': self.assertGreater}[better]
        for line in self.get_all_lines(self.resultfn):
            asserter(float(line.strip('\n').split('\t')[confcol - 1]), conflvl)

    def test_confidence_omit_confcol(self):
        options = ['--confidence-better', 'lower', '--confidence-lvl', '0.01']
        res = self.run_command(options, return_error=True)
        if res.returncode != 0:
            self.assertEqual(res.stdout.strip(), 
                    'Must define either --confcol or --confcolpattern')
        else:
            self.fail('This test should error')

    def test_omit_conf_better(self):
        options = ['--confidence-col', '1', '--confidence-lvl', '0.01']
        res = self.run_command(options, return_error=True)
        if res.returncode != 0:
            self.assertIn('usage', res.stderr.strip())
        else:
            self.fail('This test should error')

    def test_omit_conf_val(self):
        options = ['--confidence-col', '1', '--confidence-better', 'lower']
        res = self.run_command(options, return_error=True)
        if res.returncode != 0:
            self.assertIn('usage', res.stderr.strip())
        else:
            self.fail('This test should error')


class DeleteSet(MzidWithDB):
    command = 'deletesets'
    infilename = 'target_pg.tsv'
    dbfn = 'target_psms.sqlite'

    def test_deleteset(self):
        set_to_del = 'Set1'
        db = sqlite3.connect(self.workdb)
        # Add peptide which is to be deleted since it will not belong to a remaining
        # set (peptides in set1/2 in DB are identical)
        db.execute('INSERT INTO peptide_sequences(sequence) VALUES("fake_seq")')
        db.commit()
        
        rowsql = 'SELECT MIN(rownr), MAX(rownr) FROM psmrows'
        minrow, maxrow = db.execute(rowsql).fetchone()
        pepsql = 'SELECT COUNT(*) FROM peptide_sequences'
        pep_precount = db.execute(pepsql).fetchone()[0]
        exprownr = 0
        with open(self.infile) as fp:
            head = next(fp).strip('\n').split('\t')
            for line in fp:
                line = {h: l for h, l in zip(head, line.strip('\n').split('\t'))}
                if line['Biological set'] != set_to_del:
                    exprownr += 1
        self.run_command(['--dbfile', self.workdb, '--setnames', 'Set1'])
        newrownr = 0
        with open(self.resultfn) as fp:
            head = next(fp).strip('\n').split('\t')
            for line in fp:
                line = {h: l for h, l in zip(head, line.strip('\n').split('\t'))}
                self.assertNotEqual(line['Biological set'], set_to_del)
                newrownr += 1
        self.assertEqual(newrownr, exprownr)
        newmin, newmax = db.execute(rowsql).fetchone()
        self.assertEqual(minrow, newmin)
        self.assertGreater(maxrow, newmax)
        self.assertEqual(exprownr, newmax + 1)
        pep_newcount = db.execute(pepsql).fetchone()[0]
        self.assertGreater(pep_precount, pep_newcount)
        biosql = """SELECT COUNT(set_name) FROM biosets WHERE set_name='Set1'"""
        self.assertFalse(db.execute(biosql).fetchone()[0])


class TestIsoSummarize(basetests.MzidTSVBaseTest):
    """Tests producing PSM ratios, not actually summarizing"""
    suffix = '_ratio_isobaric.txt'
    command = 'isosummarize'
    infilename = 'set1_target_pg.tsv'

    def test_mediansweep(self):
        result = self.run_command(['--isobquantcolpattern', 'plex',
            '--mediansweep'])
        self.do_check(0, result.stdout, ratiomethod='sweep')

    def test_summarize_avg(self):
        denompats = ['_126']
        result = self.run_command(['--isobquantcolpattern', 'plex',
            '--denompatterns', *denompats, '--summarize-average'])
        self.do_check(0, result.stdout, denompats=denompats)

    def test_denomcolpattern_regex(self):
        denompats = ['_1[23][61]']
        result = self.run_command(['--isobquantcolpattern', 'plex', 
            '--denompatterns', *denompats])
        self.do_check(0, result.stdout, denompats=denompats)

    def get_denominator(self, line, method, denom_ch):
        if method == 'denoms':
            denomvals = [float(line[ch]) for ch in denom_ch if line[ch] != 'NA']
            if denomvals == []:
                return 0
            else:
                return sum(denomvals) / len(denomvals)
        elif method == 'sweep':
            return median([float(line[ch]) for ch in line.keys() if line[ch] != 'NA'])

    def get_infile_lines(self, infile=None):
        if infile is None:
            infile = self.infile[0]
        with open(infile) as fp:
            header = next(fp).strip('\n').split('\t')
            for line in fp:
                line = line.strip('\n').split('\t')
                yield {field: val for field, val in zip(header, line)}

    def check_normalize_medians(self, channels, denom_ch, minint, stdout,
                                medianpsms):
        ch_medians = {ch: [] for ch in channels}
        for line in self.get_infile_lines(medianpsms):
            line.update({ch: line[ch]
                         if line[ch] != 'NA' and float(line[ch]) > minint
                         else 'NA' for ch in channels})
            denom = self.get_denominator(line, denom_ch)
            if denom == 0:
                continue
            for ch in channels:
                if line[ch] == 'NA':
                    continue
                ch_medians[ch].append(float(line[ch]) / denom)
        ch_medians = {ch: median(vals) for ch, vals in ch_medians.items()}
        stdout = stdout.decode().split('\n')
        self.assertEqual(stdout[0],
                         'Channel intensity medians used for normalization:')
        stdout_channels = {x.split(' - ')[0]: x.split(' - ')[1]
                           for x in stdout[1:]}
        for ch in channels:
            self.assertEqual(float(stdout_channels[ch]), ch_medians[ch])
        return ch_medians

    def do_check(self, minint, stdout, normalize=False, medianpsms=None,
                 ratiomethod='denoms', resultch=False, denompats=False):
        channels = ['tmt10plex_126'] + [x.format('tmt10plex_1', y+27) for x in ['{}{}C', '{}{}N'] for y in range(4)] + ['tmt10plex_131']
        resultch = ['#SpecFile', 'SpecID'] + ['ratio_{}'.format(x) for x in channels]
        denom_ch = []
        if denompats:
            for denompat in denompats:
                denom_ch.extend([x for x in channels if re.search(denompat, x)])
        if normalize:
            ch_medians = self.check_normalize_medians(channels, denom_ch,
                                                      minint, stdout,
                                                      medianpsms)
        results = [x for x in self.get_values(resultch)]
        resultspecids = [f'{x[0][1]}_{x[1][1]}' for x in results]
        for line_num, in_line in enumerate(self.get_infile_lines()):
            in_line.update({ch: in_line[ch]
                            if in_line[ch] != 'NA' and
                            float(in_line[ch]) > minint else 'NA'
                            for ch in channels})
            specid = f'{in_line["#SpecFile"]}_{in_line["SpecID"]}'
            denom = self.get_denominator({ch: in_line[ch] for ch in channels},
                    ratiomethod, denom_ch)
            if denom == 0:
                exp_line = ['NA'] * len(channels)
            elif normalize:
                exp_line = [str((float(in_line[ch]) / denom) / ch_medians[ch])
                            if in_line[ch] != 'NA' else 'NA'
                            for ch in channels]
            else:
                exp_line = [str((float(in_line[ch]) / denom))
                            if in_line[ch] != 'NA' else 'NA'
                            for ch in channels]
            self.assertIn(specid, resultspecids)
            nextres = results.pop(0)
            resultline = [x[1] for x in nextres]
            self.assertEqual(resultline[2:], exp_line)




class TestIsoFeatSummarize(basetests.MzidTSVBaseTest):
    suffix = '_ratio_isobaric.txt'
    command = 'isosummarize'
    infilename = 'set1_target_pg.tsv'
    channels = ['tmt10plex_{}'.format(x) for x in ['126', '127N', '127C',
                                                   '128N', '128C', '129N',
                                                   '129C', '130N', '130C',
                                                   '131']]
    nopsms = ['{} - Quanted PSM count'.format(ch) for ch in channels]

    def test_isoquant(self):
        options = ['--featcol', '14', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126']
        self.run_command(options)
        self.isoquant_check(os.path.join(self.fixdir, 'proteins_quantonly.txt'),
            'Master protein(s)', self.channels, self.nopsms)

    def test_normalized_isoquant(self):
        options = ['--featcol', '11', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126', '--keep-psms-na-quant']
        self.run_command(options)
        self.isoquant_check(os.path.join(self.fixdir, 'isosum_charge_column.txt'), # 'proteins_isosum_column.txt'),
            'Charge', self.channels, self.nopsms)

    def test_split_multi_mapping(self):
        options = ['--featcol', '13', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126', '--split-multi-entries']
        self.run_command(options)
        self.isoquant_check(os.path.join(self.fixdir, 'proteins_quantonly_splitmulti.txt'),
            'Protein', self.channels, self.nopsms)



from app.lookups.sqlite.base import ResultLookupInterface


class ProtPepTable(ResultLookupInterface):
    table_map = {'protein': {'fntable': 'protein_tables',
                             'feattable': 'proteins',
                             'isoqtable': 'protein_iso_quanted',
                             'isochtable': 'protquant_channels',
                             'prectable': 'protein_precur_quanted',
                             'fdrtable': 'protein_fdr',
                             'fullqpsmtable': 'protein_iso_fullpsms',
                             },
                 'gene': {'fntable': 'gene_tables',
                          'feattable': 'genes',
                          'isoqtable': 'gene_iso_quanted',
                          'isochtable': 'genequant_channels',
                          'prectable': 'gene_precur_quanted',
                          'fdrtable': 'gene_fdr',
                          'fullqpsmtable': 'gene_iso_fullpsms',
                          },
                 'assoc': {'fntable': 'gene_tables',
                           'feattable': 'associated_ids',
                           'isoqtable': 'assoc_iso_quanted',
                           'isochtable': 'genequant_channels',
                           'prectable': 'assoc_precur_quanted',
                           'fdrtable': 'assoc_fdr',
                           'fullqpsmtable': 'assoc_iso_fullpsms',
                           },
                 'peptide': {'fntable': 'peptide_tables',
                             'feattable': 'peptide_sequences',
                             'isoqtable': 'peptide_iso_quanted',
                             'isochtable': 'pepquant_channels',
                             'prectable': 'peptide_precur_quanted',
                             'fdrtable': 'peptide_fdr',
                             'peptable': 'peptide_pep',
                             'fullqpsmtable': 'peptide_iso_fullpsms',
                             'flrtable': 'ptm_flr',
                             }
                 }

    def __init__(self, fn=None):
        super().__init__(fn)
        self.singlecols_to_index = []

    def get_all_poolnames(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT DISTINCT set_name, set_id FROM biosets ORDER BY set_name')
        return cursor

    def store_table_files(self, tables):
        table = self.table_map[self.datatype]['fntable']
        self.store_many(
                'INSERT INTO {}(set_id, filename) VALUES(?, ?)'.format(table), tables)
        self.index_column('table_set_ix', table, 'set_id')

    def store_quant_channels(self, quantchannels, psmnrcols):
        table = self.table_map[self.datatype]['isochtable']
        if psmnrcols:
            sql = ('INSERT INTO {}({}, channel_name, amount_psms_name) VALUES'
                   '(?, ?, ?)'.format(table, self.colmap[table][1]))
        else:
            sql = ('INSERT INTO {}({}, channel_name) VALUES'
                   '(?, ?)'.format(table, self.colmap[table][1]))
        self.store_many(sql, quantchannels)

    def get_quantchannel_map(self):
        outdict = {}
        table = self.table_map[self.datatype]['isochtable']
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT channel_id, {}, channel_name, amount_psms_name'
            ' FROM {}'.format(self.colmap[table][1], table))
        for channel_id, fnid, channel_name, amount_psms_name in cursor:
            try:
                outdict[fnid][channel_name] = (channel_id, amount_psms_name)
            except KeyError:
                outdict[fnid] = {channel_name: (channel_id, amount_psms_name)}
        return outdict

    def store_isobaric_quants(self, quants, psmnrcols):
        table = self.table_map[self.datatype]['isoqtable']
        if psmnrcols:
            sql = ('INSERT INTO {}({}, channel_id, quantvalue, amount_psms) ' 
                   'VALUES ' '(?, ?, ?, ?)'.format(table, self.colmap[table][1]))
        else:
            sql = ('INSERT INTO {}({}, channel_id, quantvalue) ' 
                   'VALUES ' '(?, ?, ?)'.format(table, self.colmap[table][1]))
        self.store_many(sql, quants)

    def index_iso(self):
        table = self.table_map[self.datatype]['isoqtable']
        self.index_column('isoq_acc_ix', table, self.colmap[table][1])
        self.index_column('isoq_ch_ix', table, self.colmap[table][2])

    def index_singlecol_data(self):
        for ixname, table, col in self.singlecols_to_index:
            self.index_column(ixname, table, col)

    def get_tablefn_map(self):
        table = self.table_map[self.datatype]['fntable']
        cursor = self.get_cursor()
        cursor.execute('SELECT * FROM {} '.format(table))
        return {fn: table_id for (table_id, setid, fn) in cursor}

    def get_feature_map(self):
        table = self.table_map[self.datatype]['feattable']
        columns = self.colmap[table][0:2]
        cursor = self.get_cursor()
        cursor.execute('SELECT {}, {} FROM {}'.format(columns[0], columns[1],
                                                      table))
        return {acc: table_id for (table_id, acc) in cursor}

    def store_singlecol(self, tablekey, vals):
        table = self.table_map[self.datatype][tablekey]
        cols = self.colmap[table]
        self.store_many('INSERT INTO {}({}, {}, {}) '
                        'VALUES (?, ?, ?)'.format(table, cols[0], cols[1],
                                                  cols[2]), vals)

    def store_precursor_quants(self, quants):
        self.store_singlecol('prectable', quants)
        table = self.table_map[self.datatype]['prectable']
        self.singlecols_to_index.append(('precursor_acc_ix', table, self.colmap[table][0]))
        self.singlecols_to_index.append(('precursor_table_ix', table, self.colmap[table][1]))

    def store_fdr(self, fdr):
        self.store_singlecol('fdrtable', fdr)
        table = self.table_map[self.datatype]['fdrtable']
        self.singlecols_to_index.append(('fdr_acc_ix', table, self.colmap[table][0]))
        self.singlecols_to_index.append(('fdr_table_ix', table, self.colmap[table][1]))

    def store_pep(self, pep):
        '''Overridden in peptide table lookup'''
        pass

    def store_fullq_psms(self, fullqpsms):
        self.store_singlecol('fullqpsmtable', fullqpsms)
        table = self.table_map[self.datatype]['fullqpsmtable']
        self.singlecols_to_index.append(('fullq_acc_ix', table, self.colmap[table][0]))
        self.singlecols_to_index.append(('fullq_table_ix', table, self.colmap[table][1]))

    def store_ptm_flr(self, flrs):
        self.store_singlecol('flrtable', flrs)
        table = self.table_map[self.datatype]['flrtable']
        self.singlecols_to_index.append(('flr_acc_ix', table, self.colmap[table][0]))
        self.singlecols_to_index.append(('flr_table_ix', table, self.colmap[table][1]))

    def check_isoquant_psmnrs(self):
        cursor = self.get_cursor()
        tables = self.table_map[self.datatype]
        psmschecksql = """SELECT EXISTS(SELECT amount_psms FROM {} 
                WHERE amount_psms IS NOT NULL)""".format(tables['isoqtable'])
        return cursor.execute(psmschecksql).fetchone()[0]

    def get_isoquant_headernames(self, stored_psmnrs):
        cursor = self.get_cursor()
        tables = self.table_map[self.datatype]
        if self.datatype == 'peptide':
            table_id_fieldname = 'peptable_id'
        else:
            table_id_fieldname = f"{tables['fntable'][:4]}table_id"
        if stored_psmnrs:
            sql = ('SELECT set_name || "_" || channel_name, set_name || "_" || amount_psms_name '
                    f'FROM {tables["isochtable"]} '
                    f'JOIN {tables["fntable"]} USING({table_id_fieldname}) '
                    'JOIN biosets USING(set_id)')
        else:
            sql = ('SELECT set_name || "_" || channel_name '
                    f'FROM {tables["isochtable"]} '
                    f'JOIN {tables["fntable"]} USING({table_id_fieldname}) '
                    'JOIN biosets USING(set_id)')
        return cursor.execute(sql)



import os
import sys
import re
from app.readers import mzidplus as readers
from app.readers import tsv as tsvreader
from app.readers import xml
from app.readers import percolator as percoreader 
from app.dataformats import mzidtsv as psmheaders
from app.dataformats import qvality as qvalityheaders


def get_scores_fdrs_from_percoqvality(percofn, qvality_psms, qvality_peps):
    '''From percolator output get svm scores, and use a separate qvality output
    for getting the calculated FDR values. Nice in case of recalculating FDR
    after filtering etc
    '''
    ns = xml.get_namespace_from_top(percofn, None)
    psms = {}
    for qvalitypsm, psm in zip(qvality_psms, percoreader.generate_psms(percofn, ns)):
        psms[psm.attrib['{%s}psm_id' % ns['xmlns']]] = {
                'svm': float(psm.find('{%s}svm_score' % ns['xmlns']).text),
                'qval': float(qvalitypsm[qvalityheaders.HEADER_QVAL]),
                'pep': float(qvalitypsm[qvalityheaders.HEADER_PEP]),
                'td': 'decoy' if psm.attrib['{%s}decoy' % ns['xmlns']] == 'true' else 'target',
                }
    for qvalitypep, peptide in zip(qvality_peps, percoreader.generate_peptides(percofn, ns)):
        psm_ids = peptide.find('{%s}psm_ids' % ns['xmlns'])
        for psm_id in psm_ids:
            try:
                psms[psm_id.text].update({
                    'pepqval': float(qvalitypep[qvalityheaders.HEADER_QVAL]),
                    'peppep': float(qvalitypep[qvalityheaders.HEADER_PEP]),
                    })
            except KeyError:
                # Edgecase, sometimes filtering percolator data causes this:
                # when there is a peptide with a PSM ID not in percolator PSMs 
                # set all its PSMs pep-q-values to 1
                [psms[x.text].update({'pepqval': 1, 'peppep': 1}) for x in psm_ids if x.text in psms]
                break
    return psms



def get_scores_fdrs_from_percoxml(percofn):
    """
    From a percolator XML output file, take the FDR (q-value) and PEP and apply
    to the PSM table
    """
    ns = xml.get_namespace_from_top(percofn, None)
    # Get PSM FDRs
    psms = {
            p.attrib['{%s}psm_id' % ns['xmlns']]: {
                'svm': float(p.find('{%s}svm_score' % ns['xmlns']).text),
                'qval': float(p.find('{%s}q_value' % ns['xmlns']).text),
                'pep': float(p.find('{%s}pep' % ns['xmlns']).text),
                'td': 'decoy' if p.attrib['{%s}decoy' % ns['xmlns']] == 'true' else 'target',
                }
            for p in percoreader.generate_psms(percofn, ns)
            }
    # Then get peptide FDRs and add them to PSMs
    for peptide in percoreader.generate_peptides(percofn, ns):
        psm_ids = peptide.find('{%s}psm_ids' % ns['xmlns'])
        for psm_id in psm_ids:
            try:
                psms[psm_id.text].update({
                    'pepqval': float(peptide.find('{%s}q_value' % ns['xmlns']).text),
                    'peppep': float(peptide.find('{%s}pep' % ns['xmlns']).text),
                    })
            except KeyError:
                # Edgecase, sometimes filtering percolator data causes this:
                # when there is a peptide with a PSM ID not in percolator PSMs 
                # set all its PSMs pep-q-values to 1
                [psms[x.text].update({'pepqval': 1, 'peppep': 1}) for x in psm_ids if x.text in psms]
                break
    return psms


def generate_spec_id_items(mzid_specidr, mzns):
    scan = 0
    has_scans = True
    for specidr in mzid_specidr:
        if has_scans:
            try:
                scan = int({x.split('=')[0]: x.split('=')[1] for x in specidr.attrib['spectrumID'].split(' ')}['scan'])
            except KeyError:
                # in e.g. timstof data there are no true scan numbers, percolator sets it by increment
                has_scans = False
                scan += 1
        else:
            scan += 1
        for specidi in specidr.findall('{%s}SpectrumIdentificationItem' % mzns['xmlns']):
            yield scan, specidi


def add_fdr_to_mzidtsv(psms, mzid_specidr, mzns, percodata):
    """Takes PSMs from an mzIdentML and its MSGF+ TSV and a corresponding 
    percolator XML. 
    """
    # mzId results and PSM lines can be zipped
    # but we assume that multi-solution scans (e.g. Isoleucine/leucine)
    # have identical scoring, i.e. MSGF is run with -n1 (default)

    specidis = generate_spec_id_items(mzid_specidr, mzns)
    for psm, (scan, specidi) in zip(psms, specidis):
        spfile = os.path.splitext(psm[psmheaders.HEADER_SPECFILE])[0]
        psmcharge = psm[psmheaders.HEADER_CHARGE]
        perco_id = f'{spfile}_{specidi.attrib["id"]}_{scan}_{psmcharge}_{specidi.attrib["rank"]}'
        try:
            percopsm = percodata[perco_id]
        except KeyError:
            continue
        outpsm = {k: v for k,v in psm.items()}
        psmscores = {
            psmheaders.HEADER_SVMSCORE: percopsm['svm'], 
            psmheaders.HEADER_PSMQ: percopsm['qval'], 
            psmheaders.HEADER_PSM_PEP: percopsm['pep'], 
            }
        try:
            pepscores = {
                psmheaders.HEADER_PEPTIDE_Q: percopsm['pepqval'], 
                psmheaders.HEADER_PEPTIDE_PEP: percopsm['peppep'], 
                psmheaders.HEADER_TARGETDECOY: percopsm['td'],
                }
        except KeyError:
            pepscores = {
                psmheaders.HEADER_PEPTIDE_Q: 1,
                psmheaders.HEADER_PEPTIDE_PEP: 1,
                psmheaders.HEADER_TARGETDECOY: 1,
                }
        outpsm.update({**psmscores, **pepscores})
        # Remove all decoy protein matches from target proteins, to ensure downstream
        # processing does not trip up on them, e.g. having a decoy master protein.
        if percopsm['td'] != 'decoy':
            outprots = []
            for prot in outpsm[psmheaders.HEADER_PROTEIN].split(';'):
                if not prot.startswith(psmheaders.DECOY_PREFIX):
                    outprots.append(prot)
            outpsm[psmheaders.HEADER_PROTEIN] = ';'.join(outprots)
        yield outpsm


def get_header_with_percolator(oldheader):
    ix = oldheader.index(psmheaders.HEADER_EVALUE) + 1
    return oldheader[:ix] + psmheaders.PERCO_HEADER + oldheader[ix:]



import sys
import re
from os import environ
from Bio import SeqIO

from string import Template

from create_modfile import get_msgfmods, categorize_mod, parse_cmd_mod, modpos, NON_BLOCKING_MODS

from dataformats import mzidtsv as dm

def get_msgf_seq_mass(mass):
    return '{}{}'.format('+' if mass > 0 else '', str(round(mass, 3)))


def get_luci_mod(pos, aa, mass):
    if pos == 'N-term':
        residue = '['
    elif pos == 'C-term':
        residue = ']'
    else:
        residue = aa
    return '{} {}'.format(residue, mass)


def get_ptm_mass(ptmnames, ):
    ptms_mass = []
    for cmdptm in ptmnames:
        for ptm in parse_cmd_mod(cmdptm, msgfmods):
            # FIXME OPTIONAL competition for multi-residue/line spec, how to know which 
            # modifications can compete?
            ptm = ptm.split(',')
            realmodmass = ptm[0]
            if modpos(ptm) in fixedmods:
                fixmass = sum([float(x[0]) for x in fixedmods[modpos(ptm)]])
                ptm[0] = str(round(-(fixmass - float(ptm[0])), 5))
                print(realmodmass, ptm)
            ptms_mass.append(add_mods_translationtable(ptm, massconversion_msgf))
            for residue in ptm[1]:
                luciptm = get_luci_mod([ptm[0], residue, *ptm[2:]])
                ## TODO
                ## Now we add e.g. -187 if acetyl mod competes with fixed TMT
                ## That does not work, but the below line also doesnt work, which
                ## Puts the actual +42 mass in the luciphor config
                ## Total mass of residue is in the luci input K+42 = 170
                ## Luciphor ignores fixed mod containing residues I think
                #luciptm = get_luci_mod([realmodmass, residue, *ptm[2:]])
                if luciptm not in target_mods:
                    target_mods.append(luciptm)
                    decoy_mods.add(ptm[0])
        # Neutral loss only for phospho (and possibly glycosylation), add more if we need
        nlosses, decoy_nloss = [], []
        if ptm[4] == 'Phospho':
            nlosses.append('sty -H3PO4 -97.97690')
            decoy_nloss.append('X -H3PO4 -97.07690')
    return ptms_mass

###############
# FIXME we can also incorporate create_modfile for MSGF

def create_luciphor_input(psms, template, fixedmods, varmods, ptms, ms2tol, ms2toltype, mzmlpath,
       lucialgo, maxcharge, maxlen, minpsms, threads):
    """Writes the following files:
        - luciphor_config_input - the config file
        - luciphor_psm_input - the PSMs with PTM in luciphor format
        - ptm_psms - the PSMs for adding luciphor to afterwards

    Fixed mods, var mods, ptms - {'mass': float, 'aa': str,  'name': str, 'pos': str}
    """
    massconversion_msgf = {}
    target_mods, decoy_mods, lucifixed, lucivar = [], set(), [], []

    # Prep fixed mods for luciphor template
    for mod in fixedmods:
        massconversion_msgf[get_msgf_seq_mass(mod['masschange'])] = mod['masschange']
        lucimod = get_luci_mod(mod['pos'], mod['aa'], mod['masschange'])
        if lucimod not in lucifixed:
            lucifixed.append(lucimod)
    # Var mods too, and add to mass list to filter PSMs on later (all var mods must be annotated on sequence input)
    varmods_mass = []
    for mod in varmods:
        msgf_mass = get_msgf_seq_mass(mod['masschange'])
        massconversion_msgf[msgf_mass] = mod['masschange']
        varmods_mass.append(msgf_mass)
        lucimod = get_luci_mod(mod['pos'], mod['aa'], mod['masschange'])
        if lucimod not in lucivar:
            lucivar.append(lucimod)

    # Get PTMs from cmd line and prep for template
    ptms_mass = set()
    nlosses, decoy_nloss = set(), set()
    phos_nloss = '-H3PO4 -97.97690'
    for ptm in ptms:
        luciptm = get_luci_mod(ptm['pos'], ptm['aa'], ptm['masschange'])
        if luciptm not in target_mods:
            target_mods.append(luciptm)
            decoy_mods.add(ptm['masschange'])
        msgf_modmass = get_msgf_seq_mass(ptm['masschange'])
        massconversion_msgf[msgf_modmass] = ptm['masschange']
        ptms_mass.add(msgf_modmass)
        # Neutral loss only for phospho (and possibly glycosylation), add more if we need
        if ptm['name'] == 'Phospho':
            nlosses.add(f'{ptm['aa']} {phos_nloss}')
            decoy_nloss.add(f'X {phos_nloss}')

    # Create the config file contents
    with open(template) as fp:
        lucitemplate = Template(fp.read())
    lt_render = lucitemplate.substitute(mzml_path=mzml_path, algo=lucialgo, maxcharge=maxcharge,
            maxpeplen=maxlen, minpsms=minpsms, outfile='luciphor.out', 
            ms2tol=ms2tol,
            ms2toltype=ms2toltype,
            dmasses=decoy_mods,
            neutralloss=nlosses,
            decoy_nloss=decoy_nloss
            thread=threads
            ))
    for tmod in lucifixed:
        lt_render = f'{lt_render}\nFIXED_MOD = {tmod}'
    for tmod in lucivar:
        lt_render = f'{lt_render}\nVAR_MOD = {tmod}'
    for tmod in target_mods:
        lt_render = f'{lt_render}\nVAR_MOD = {tmod}'
    for tmod in nlosses:
        lt_render = f'{lt_render}\nNL = {tmod}'
    for tmod in decoy_nloss:
        lt_render = f'{lt_render}\nDECOY_NL = {tmod}'
    for tmod in decoy_mods:
        lt_render = f'{lt_render}\nDECOY_MASS = {tmod}'
    yield lt_render

    # TODO
    # How to handle competing mods? I have yet to find a use-case but theoretically
    # a fixed/PTM competition may occur (not Acetyl/TMT since Ac is not labile like Phos)
    # Probably replace double notation (tmt - (tmt - mod)), eg 229-187 in PSM table with 
    # the actual PTM mass (42)? But luciphor ignores PTMs on residues with a fixed mod anyway

    # Create the Lucxor PSM input table
    for psm in psms:
        outpsm = {k: v for k,v in psm.items()}
        ptm_in_seq = False
        sites = []
        pep, start = '', 0
        seq = psm[dm.HEADER_PEPTIDE]
        # TODO below mod notation is MSGF+ dependent, may need changing
        for mod in re.finditer('[\+\-0-9]+.[\+\-.0-9]+', seq):
            modtxt = mod.group()
            if modtxt.count('+') + modtxt.count('-') > 1:
                # multi mod on single residue
                multimods = re.findall('[\+\-][0-9.]+', modtxt)
            else:
                multimods = [modtxt]
            modmass = sum([massconversion_msgf[x] for x in multimods])
            if set(multimods).intersection(ptms_mass):
                ptm_in_seq = True
                # IF fixed + target -> use the target mass from the msgfmods file
                # target + target -> 
            pep += seq[start:mod.start()]
            if not set(multimods).intersection([*ptms_mass, *varmods_mass]):
                # Do not annotate fixed mods
                start = mod.end()
                continue
            if len(pep) == 0:
                aamass = modmass
            else:
                aamass = round(dm.AA_WEIGHTS_MONO[pep[-1]] + modmass, 5)
            sites.append([len(pep)- 1, aamass])
            start = mod.end()
        if not ptm_in_seq:
            # only output PSMs with PTM for luciphor
            continue
        if start != mod.endpos:
            pep += seq[start:]
        if sites[0][0] == -1: sites[0][0] = -100
        # TODO add C-terminal mods (rare)?
        outpsm.update({
            'peptide': pep,
            'modSites': ','.join(['{}={}'.format(x[0], x[1]) for x in sites])})
        yield outpsm


def parse_luciphor_output(luciptms, luciscores, outpsms, tdb, minscore_high, ptmmods):
    modresidues, ptmmasses = {}, {}
    for mod in ptmmods:
        modresidues[mod['name']] = []
        totalmass = mod['mass'] if mod['aa'] == '*' else dm.AA_WEIGHTS_MONO[mod['aa']] + mod['mass']]
        ptmmasses[round(totalmass, 0)] = modname
        
    # Now go through scores, luciphor and PSM table
#    with open('all_scores.debug') as scorefp, open('luciphor.out') as fp, open('psms') as psms, open(outfile, 'w') as wfp:
#        outheader = psmheader + PTMFIELDS

    scorepep = {'specId': False}
    lucptms = {}
    for lucptm in luciptms:
        out = {}
        specid = lucptm['specId']
        ptm = {
            TOPFLR: lucptm['globalFLR'],
            TOPSCORE: lucptm['pep1score'],
            OTHERPTMS: 'NA',
            }
        # match the modified residues and group:
        barepep, start = '', 0
        modpep = lucptm['predictedPep1']
        for x in re.finditer('([A-Z]){0,1}\[([0-9]+)\]', modpep):
            if x.group(1) is not None: # check if residue (or protein N-term)
                barepep += modpep[start:x.start()+1]
            else:
                barepep += modpep[start:x.start()]
            start = x.end()
            modresidues[ptmmasses[int(x.group(2))]].append((x.group(1), len(barepep)))

        # modresidues = { ptmname: [('S', 13), ('S', 22), ...], ...}
        barepep += modpep[start:]
        ptm.update({'barepep': barepep, 'modres': modresidues})
        ptm[TOPPTM] = ';'.join(['{}:{}'.format(name, ','.join(['{}{}'.format(x[0], x[1]) for x in resmods])) for 
                name, resmods in modresidues.items() if len(resmods)])

        # Get other highscoring permutations from all_scores file
        extrapeps = []
        if scorepep['specId'] == specid:
            lucptm['likeScoredPep'] = re.sub(r'([A-Z])\[[0-9]+\]', lambda x: x.group(1).lower(), lucptm['predictedPep1'])
            if scorepep['curPermutation'] != lucptm['likeScoredPep'] and float(scorepep['score']) > minscore_high:
                extrapeps.append('{}:{}'.format(','.join(['{}{}'.format(x.group().upper(), x.start() + 1)
                            for x in re.finditer('[a-z]', scorepep['curPermutation'])]),
                            scorepep['score']))
        for scorepep in luciscores:
            if int(scorepep['isDecoy']):
                continue
            if scorepep['specId'] != specid:
                break
            lucptm['likeScoredPep'] = re.sub(r'([A-Z])\[[0-9]+\]', lambda x: x.group(1).lower(), lucptm['predictedPep1'])
            if scorepep['curPermutation'] != lucptm['likeScoredPep'] and float(scorepep['score']) > minscore_high:
                extrapeps.append('{}:{}'.format(','.join(['{}{}'.format(x.group().upper(), x.start() + 1)
                        for x in re.finditer('[a-z]', scorepep['curPermutation'])]),
                        scorepep['score']))
            if len(extrapeps):
                ptm[OTHERPTMS] = ';'.join(extrapeps)
        lucptms[specid] = ptm

    # Now output to its PSM table
    for psm in outpsms:
        spfile = os.path.splitext(psm[dm.HEADER_SPECFILE])[0]
        luc_psmid = f'{spfile}.{psm[dm.HEADER_SCANNR]}.{psm[dm.HEADER_SCANNR]}.{dm.HEADER_CHARGE}'
        if luc_psmid in lucptms:
            ptm = lucptms[luc_psmid]
            # Get protein site location of mod
            if dm.HEADER_MASTER_PROT in psm:
                proteins = psm[dm.HEADER_MASTER_PROT].split(';')
                proteins = {p: tdb[p].seq.find(ptm['barepep']) for p in proteins}
                proteins_loc = {p: [] for p, peploc in proteins.items() if peploc > -1}
                for p, peploc in proteins.items():
                    for ptmname, ptmlocs in ptm['modres'].items():
                        if ptmname not in ptms:
                            continue
                        protptms = []
                        for res_loc in ptmlocs:
                            protptms.append('{}{}'.format(res_loc[0], res_loc[1] + peploc))
                            proteins_loc[p].append('{}_{}'.format(ptmname, ','.join(protptms)))
                psm[dm.HEADER_MASTER_PROT] = ';'.join(['{}:{}'.format(p, ':'.join(ptmloc)) for p, ptmloc in proteins_loc.items()])
            outpsm = {k: v for k,v in psm.items()}
            outpsm.update(ptm)
            outpsm[SE_PEPTIDE] = outpsm.pop(PEPTIDE)
            outpsm[PEPTIDE] = '{}_{}'.format(ptm['barepep'], ptm[TOPPTM])
            yield outpsm



import argparse

shared_options = {
    'fn': {'driverattr': 'fn', 'type': 'file', 'clarg': '-i',
           'help': 'Input file of {} format'},
    'outfile': {'driverattr': 'outfile', 'type': str,
                'clarg': '-o', 'help': 'Output file', 'required': False},
    'outdir': {'driverattr': 'outdir', 'clarg': '-d',
               'help': 'Directory to output in', 'type': 'file',
               'required': False},
    'multifiles': {'driverattr': 'fn', 'clarg': '-i',
                   'help': 'Multiple input files of {} format',
                   'type': 'file', 'nargs': '+'},
    'lookupfn': {'driverattr': 'lookupfn', 'clarg': '--dbfile',
                 'type': 'file', 'help': 'Database lookup file'},
    'setnames': {'driverattr': 'setnames',
                 'type': str, 'nargs': '+', 'clarg': '--setnames',
                 'help': 'Names of biological sets. Can be '
                 'specified with quotation marks if spaces are '
                 'used'},
    'spectracol': {'driverattr': 'spectracol',
                   'type': int, 'clarg': '--spectracol', 'help':
                   'Column number in which spectra file names are, '
                   'in case some framework has changed the file '
                   'names. First column number is 1.', 'required': False,
                   'default': 1},
    'proline': {'driverattr': 'proline', 'required': False,
                'clarg': '--cutproline', 'action': 'store_const',
                'const': True, 'default': False, 'help': 'Flag to make '
                'trypsin before a proline residue. Then filtering will be '
                'done against both cut and non-cut peptides.',
                },
    'fasta': {'driverattr': 'fasta',
              'type': 'file', 'help': 'FASTA sequence database',
              'required': False, 'default': False, 'clarg': '--fasta'},
    'unroll': {'driverattr': 'unroll', 'clarg': '--unroll', 'const': True,
               'action': 'store_const', 'default': False, 'help': 'PSM table '
               'from Mzid2TSV contains either one PSM per line with all '
               'the proteins of that shared peptide on the same line (not'
               ' unrolled, default), or one PSM/protein match per line '
               'where each protein from that shared peptide gets its own '
               'line (unrolled).', 'required': False},
    'isobaric': {'driverattr': 'isobaric', 'clarg': '--isobaric',
                 'action': 'store_const', 'const': True, 'default': False,
                 'help': 'Specifies to add isobaric quant data from lookup DB '
                 'to output table', 'required': False,
                 },
    'precursor': {'driverattr': 'precursor', 'clarg': '--ms1quant',
                  'action': 'store_const', 'const': True, 'default': False,
                  'help': 'Specifies to add precursor quant data from lookup '
                  'DB to output table', 'required': False,
                  },
    'quantcolpattern': {'driverattr': 'quantcolpattern',
                        'clarg': '--isobquantcolpattern', 'type': str,
                        'default': False, 'required': False,
                        'help': 'Unique text pattern to identify '
                        'isobaric quant columns in input table.'},
    'precursorquantcolpattern': {'driverattr': 'precursorquantcolpattern',
                                 'type': str, 'required': False,
                                 'clarg': '--ms1quantcolpattern',
                                 'default': None,
                                 'help': 'Unique text pattern to identify '
                                 'precursor quant column in input table.'},
    'quantacccolpattern': {'driverattr': 'quantacccolpattern',
                           'clarg': '--qaccpattern', 'type': str,
                           'help': 'Unique text pattern to identify '
                           'accession column in table containing quant info.'},
    'featcol': {'driverattr': 'featcol', 'clarg': '--featcol',
                   'type': int, 'required': False, 'help': 'Column number in '
                   'table in which desired accessions are. '
                   'stored. First column number is 1. Use in case of not '
                   'using default column {}'},
    'featcolpattern': {'driverattr': 'featcolpattern', 'clarg': '--featcolpattern',
                    'type': str, 'required': False, 'help': 'Text pattern to '
                    'identify column in table in which protein or gene '
                    'accessions are. Use in case of not using standard '
                    '{} column', 'default': False},
    'fdrcolpattern': {'driverattr': 'fdrcolpattern', 'required': False,
                      'clarg': '--fdrcolpattern', 'type': str,
                      'default': None, 'help': 'Unique text pattern to identify '
                      'FDR column in input table.'},
    'flrcolpattern': {'driverattr': 'flrcolpattern', 'required': False,
                      'clarg': '--flrcolpattern', 'type': str,
                      'default': None, 'help': 'Unique text pattern to identify '
                      'FLR (peptide PTM false localization rate) column in input table.'},
    'fastadelim': {'driverattr': 'fastadelim', 'clarg': '--fastadelim',
                   'required': False, 'type': str,
                   'choices': ['tab', 'pipe', 'semicolon'],
                   'help': 'Delimiter in FASTA header, used to parse gene '
                   'names in case of non-ENSEMBL/Uniprot'},
    'genefield': {'driverattr': 'genefield', 'clarg': '--genefield',
                  'required': False, 'type': int,
                  'help': 'Field nr (first=1) in FASTA that contains gene '
                  'name when using --fastadelim to parse the gene names'},
    'minlength': {'driverattr': 'minlength', 'default': 0,
                  'help': 'Minimum length of peptide to be included',
                  'type': int, 'clarg': '--minlen', 'required': False},
    'addbioset': {'driverattr': 'addbioset',
            'clarg': '--addbioset', 'required': False, 'action': 'store_const',
            'default': False, 'const': True,
            'help': 'Add biological setname from DB lookup to PSM table',
               },
    'addmiscleav': {'driverattr': 'addmiscleav',
            'clarg': '--addmiscleav', 'required': False, 'action': 'store_const',
            'default': False, 'const': True, 'help': 'Add missed cleavages to PSM table',
               },
    'ntermmloss': {'driverattr': 'ntermmloss', 'clarg': '--nterm-meth-loss', 
        'required': False, 'action': 'store_const', 'const': True, 'default': False,
        'help': 'Include peptides in trypsinization where the protein N-term methionine '
        'residue has been lost. When used in storeseq, the filter database should '
        'be built with this flag in order for the filtering to work.' 
        },
    'fullprotein': {'driverattr': 'fullprotein', 'clarg': '--fullprotein',
        'default': False, 'action': 'store_const', 'const': True, 'help':
        'Store (storeseq) or use (seqfilt, seqmatch) full protein sequences '
        '(at a minimum-match length) in the SQLite file rather than tryptic sequences',
        'required': False},
    'minint': {'driverattr': 'minint', 'clarg': '--minint', 'type': float,
               'help': 'Intensity threshold of PSMs when calculating '
               'isobaric ratios. Values below threshold will be set to NA. '
               'Defaults to no threshold.', 'required': False, 'default': 0,
               },
    'denompatterns': {'driverattr': 'denompatterns', 'required': False,
                      'clarg': '--denompatterns', 'type': str, 'nargs': '+',
                      'help': 'Regex patterns to detect denominator channels '
                      'in a PSM table, when creating a table with summarized '
                      'feature isobaric ratios. '
                      'If both --denompatterns and --denomcol are given then column '
                      'numbers are used. Usage e.g. --denompattern _126 _131. '
                      'Also possible: --denompattern _12[6-7] to detect '
                      'multiple columns.'
                      },
    'denomcols': {'driverattr': 'denomcols', 'clarg': '--denomcols',
                  'type': int, 'nargs': '+', 'required': False,
                  'help': 'Column numbers of denominator channels when '
                  'creating a summarized feature table with isobaric ratios '
                  'from PSMs'
                  },
    'mediansweep': {'driverattr': 'mediansweep', 'clarg': '--mediansweep',
            'action': 'store_const', 'const': True, 'default': False,
            'help': 'Instead of choosing a denominator channel, use the median intensity of each '
            'PSM as its denominator.', 'required': False,
            },
    'medianintensity': {'driverattr': 'medianintensity', 'clarg': '--medianintensity',
            'action': 'store_const', 'const': True, 'default': False,
            'help': 'Instead of choosing a denominator channel or median-sweeping, '
            'report the the median intensity of each summarized feat per channel. This '
            'results in reported intensities rather than ratios.', 'required': False,
            },
    'median_or_avg': {'driverattr': 'median_or_avg', 'clarg': '--summarize-average',
            'action': 'store_const', 'const': 'average', 'default': 'median',
            'required': False, 'help': 'Use average isobaric quantification values for summarizing '
            'quant from PSMs, instead of default PSM median values'},
    'keep_psms_na': {'driverattr': 'keepnapsms', 'clarg': '--keep-psms-na-quant',
            'action': 'store_const', 'const': True, 'default': False, 'required': False,
            'help': 'When summarizing isobaric quantification data, also use '
            'the PSMs that have an NA in any channel, even if these may contain '
            'overly noisy quant data in the other channels. Normally these PSMs '
            'would be skipped in quantification'},
    'splitmultientries': {'driverattr': 'splitmultientries', 'clarg': '--split-multi-entries',
            'action': 'store_const', 'const': True, 'default': False, 'required': False,
            'help': 'When summarizing isobaric quantification data, also use '
            'the PSMs that have multiple mappings to their column, and split these, '
            'so that quant data from a PSM mapping to two proteins is used for both '
            'proteins Normally these PSMs accessionould be skipped in quantification'},
    'logisoquant': {'driverattr': 'logisoquant', 'clarg': '--logisoquant',
        'required': False, 'action': 'store_const', 'const': True, 'help':
        'Output log2 values for isoquant ratios. This log2-transforms input PSM data '
        'prior to summarizing and optional normalization. Ratios will '
        'be calculated subtracted rather than divided, obviously.'},
    'mediannormalize': {'driverattr': 'mediannormalize',
        'clarg': '--median-normalize', 'default': False, 
        'required': False, 'action': 'store_const', 'const': True,
        'help': 'Median-centering normalization for isobaric quant data on protein or '
        'peptide level. This median-centers the data for each channel by '
        'dividing with the median channel value (or subtracting in case of '
        'log data), for each channel in output features, e.g. proteins.'},
    'mednorm_factors': {'driverattr': 'mednorm_factors',
        'clarg': '--normalization-factors-table', 'type': 'file', 'required': False, 
        'help': 'A protein/peptide/gene table that provides an alternate source '
        'of normalization factors than the table to be normalized. This can be '
        'used e.g. when having a PTM table which does not have a large amount '
        'of peptides or is noisy. Use together with --median-normalize',
        },
    'ms1mediannormalize': {'driverattr': 'ms1mediannormalize',
        'clarg': '--median-normalize-ms1', 'default': False, 
        'required': False, 'action': 'store_const', 'const': True,
        'help': 'Median-centering normalization for MS1 quant data on protein or '
        'peptide level. This median-centers the data by dividing output MS1 quant '
        'values with the median output MS1 quant value '},
}

sequence_options = {
    'scramble': {
        'driverattr': 'scramble', 'clarg': '--scramble',
        'help': 'Decoy scrambling method, use: '
        '"tryp_rev": tryptic reverse, or "prot_rev": full (protein) reverse.',
        'required': False, 'default': 'tryp_rev'},
    'ignoretarget': {
        'driverattr': 'ignoretarget', 'clarg': '--ignore-target-hits',
        'help': 'Do not remove tryptic peptides from sequence where they match target DB',
        'required': False, 'action': 'store_const', 'const': True, 'default': False},
    'trypsinize': {'driverattr': 'trypsinize',
               'clarg': '--notrypsin', 'required': False,
               'action': 'store_const', 'const': False, 'default': True,
               'help': 'Do not trypsinize. In case of using a pretrypsinized FASTA file'
               },
    'keep_target': {'driverattr': 'keep_target', 'clarg': '--keep-target', 'required': False,
        'action': 'store_true', 'default': False, 'help': 'If this flag is passed, sequences '
        'that are shuffled the --max-shuffle amount of times will be kept as unshuffled '
        'tryptic reversed peptides. In case you want to keep the target/decoy DB the same '
        'sizes'},
    'max_shuffle': {'driverattr': 'max_shuffle',
               'clarg': '--maxshuffle', 'required': False, 'type': int, 'default': 10,
               'help': 'Amount of times to attempt to shuffle a decoy reversed peptide '
               'to make it not match target peptides, before discarding it.'
               ' Used when using tryptic peptide reversal (not protein reversal)'},
    'miss_cleavage': {'driverattr': 'miss_cleavage',
               'clarg': '--miscleav', 'required': False, 'type': int, 'default': 0,
               'help': 'Amount of missed cleavages to allow when trypsinizing, '
               'default is 0',
               },
    'stopcodonsplit': {'driverattr': 'stopcodonsplit', 'clarg': '--ignore-stop-codons',
        'required': False, 'action': 'store_const', 'const': False, 'default': True, 
        'help': 'Ignore stop codons in protein sequences when trypsinizing them (default is '
        'to split on them.'},
    }


lookup_options = {
    # NB this is a flag, in percolator/PSM table filtering it is an int param
    'insourcefrag': {'driverattr': 'insourcefrag',
                'clarg': '--insourcefrag', 'default': False,
                'action': 'store_const', 'const': True, 'help': 'Apply '
                'filter against both intact peptides and those '
                'that match to the C-terminal part of a tryptic peptide '
                'from the database, resulting from in-source fragmentation, '
                'where some amino acids will be missing from the N-terminus. '
                'Specify the max number of amino acids that may be missing. '
                'Database should be built with this '
                'flag in order for the lookup to work, since sequences '
                'will be stored and looked up reversed', 'required': False
                },
    'spectrafns': {'driverattr': 'spectrafns',
                   'type': str, 'help': 'Spectra files in mzML '
                   'format. Multiple files can be specified, if '
                   'order is important, e.g. when matching them '
                   'with quant data, the order will be their input '
                   'order at the command line.', 'clarg': '--spectra',
                   'nargs': '+'},
    'isobaric': {'driverattr': 'isobaricfns', 'clarg': '--isobaric', 
        'type': str, 'nargs': '+', 'required': False,
        'help': 'Isobaric quant output files from '
        'OpenMS in consensusXML '
        'format. Multiple files can be specified, '
        'and matching order with spectra files is important.',
        },
    'kronik': {'driverattr': 'kronikfns', 'clarg': '--kronik', 
        'type': str, 'nargs': '+', 'required': False,
        'help': 'MS1 persisting peptide quant output files from Kronik in text format.'
        'Multiple files can be specified, '
        'and matching order with spectra files is important.',
        },
    'dinosaur': {'driverattr': 'dinosaurfns', 'clarg': '--dinosaur', 
        'type': str, 'nargs': '+', 'required': False,
        'help': 'MS1 persisting peptide output files from Dinosaur in text format.'
        'Multiple files can be specified, '
        'and matching order with spectra files is important.',
        },
    'sum_or_apex': {'driverattr': 'sum_or_apex', 'clarg': '--apex',
        'action': 'store_const', 'const': 'apex', 'default': 'sum',
        'required': False, 'help': 'Use MS1 peak envelope apex instead of '
        'peak sum when storing quant data.'},
    'rttol': {'driverattr': 'rt_tol', 'clarg': '--rttol',
        'conditional_required': ['kronik'], 'type': float, 
        'help': 'Specifies tolerance in seconds for '
        'retention time when mapping MS1 feature quant info to '
        'identifications in the PSM table.'},
    'mztol': {'driverattr': 'mz_tol', 'clarg': '--mztol',
              'type': float, 'help': 'Specifies tolerance in mass-to-charge '
              'when mapping MS1 feature quant info to identifications in '
              'the PSM table.', 'conditional_required': ['kronik']},
    'mztoltype': {'driverattr': 'mz_toltype', 
                  'type': str, 'choices': ['ppm', 'Da'],
                  'clarg': '--mztoltype', 'conditional_required': ['kronik'],
                  'help': 'Type of tolerance in mass-to-charge when mapping '
                  'MS1 feature quant info to identifications in the PSM table.'
                  ' One of ppm, Da.'},
    'pepprotmapping': {'driverattr': 'pepprotmapping', 'clarg': '--map-accessions',
        'required': False, 'action': 'store_const', 'const': True, 'default': False,
        'help': 'Use this flag when building a peptide lookup (not --fullprotein) where '
        'you want to keep the protein mapping for later use in e.g. annotating PSM table '
        'with sequence hits to the passed FASTA'},
    'nogroup': {'driverattr': 'nogroup',
                   'clarg': '--no-group-annotation', 'action': 'store_const',
                   'const': True, 'default': False, 'required': False,
                   'help': 'For protein/peptide table merging. Do not include protein group '
                   'or gene data in output, just use protein accessions.'},
}
lookup_options['featcol'] = {k: v for k, v 
        in shared_options['featcol'].items()}
lookup_options['featcol'].update(
        {'default': 1, 'help': lookup_options['featcol']['help'].format('1')})

lookup_options['fasta'] = {k: v for k, v in shared_options['fasta'].items()}
lookup_options['fasta']['help'] = ('FASTA sequence database to use when '
                                     'extracting gene names to the PSM '
                                     'table from proteins')

percolator_options = {
    'protheaders': {'driverattr': 'protheaders', 'clarg': '--protheaders',
                    'nargs': '+', 'type': str,
                    'help': 'Specify protein FASTA headers to split on. '
                    'Multiple headers of the same split-type can be grouped '
                    'with semicolons. E.g. --protheaders \'ENSP;sp '
                    'PSEUDOGEN;ncRNA\' would split into ENSP/swissprot peptides '
                    'and pseudogenes/non-coding RNA peptides.'},
    'deamidate': {'driverattr': 'deamidate', 'clarg': '--deamidate',
                  'action': 'store_const', 'default': False, 'const': True,
                  'help': 'Filter against both normal peptides and deamidated '
                  'peptides where a D->N transition has occurred.',
                  'required': False},
    'forcetryp': {'driverattr': 'forcetryp', 'clarg': '--enforce-tryptic',
                  'action': 'store_const', 'default': False, 'const': True,
                  'help': 'When filtering peptides against whole proteins, '
                  'filter out peptides that match a whole protein but only '
                  'if they are fully tryptic, i.e. the protein needs K,R 1 '
                  'position upstream of the peptide, and the peptide '
                  'C-terminal should also be K,R. Useful when discerning '
                  'from pseudogenes', 'required': False},
    'insourcefrag': {'driverattr': 'insourcefrag',
                'clarg': '--insourcefrag',
                'type': int, 'default': 0, 'help': 'Apply '
                'filter against both intact peptides and those '
                'that match to the C-terminal part of a tryptic peptide '
                'from the database, resulting from in-source fragmentation, '
                'where some amino acids will be missing from the N-terminus. '
                'Specify the max number of amino acids that may be missing. '
                'Database should be built with this '
                'flag in order for the lookup to work, since sequences '
                'will be stored and looked up reversed', 'required': False
                },
    'deduppsms': {'driverattr': 'deduppsms', 'clarg': '--includepsms', 'const': True,
        'default': False, 'action': 'store_const', 'help': 'When running dedupperco '
        'also remove "duplicate" PSMs (by PSM ID plus sequence). Keeps first PSM '
        'encountered of each PSM ID / sequence combination', 'required': False},
}

psmtable_options = {
    'genes': {'driverattr': 'genes', 'clarg': '--genes', 'action': 'store_const', 
        'const': True, 'default': False, 'help': 'Specifies to add genes to PSM '
        'table', 'required': False,
        },
    'proteingroup': {'driverattr': 'proteingroup', 'clarg': '--proteingroup',
        'action': 'store_const', 'const': True, 'default': False, 
        'help': 'Specifies to add protein groups to PSM table', 'required': False,
        },
    'oldpsmfile': {'driverattr': 'oldpsmfile', 'clarg': '--oldpsms', 'type': 'file',
                'help': 'PSM table file containing previously analysed PSMs to '
                'append new PSM table to.', 'required': False},
    'filtpep': {'driverattr': 'filtpep', 'clarg': '--filtpep',
                'help': 'Peptide q-value cutoff level as a floating point number',
                'type': float, 'required': False},
    'filtpsm': {'driverattr': 'filtpsm', 'clarg': '--filtpsm',
                'help': 'PSM q-value cutoff level as a floating point number',
                'type': float, 'required': False},
    'confcol': {'driverattr': 'confcol', 'clarg': '--confidence-col',
                'help': 'Confidence column number or name in the tsv file. '
                'First column has number 1.', 'type': int, 'required': False},
    'confpattern': {'driverattr': 'confpattern', 'clarg': '--confcolpattern',
                    'type': str, 'required': False, 'help': 'Text pattern to '
                    'identify column in table on which confidence filtering '
                    'should be done. Use in case of not using standard '
                    '{} column', 'default': False},
    'conflvl': {'driverattr': 'conflvl', 'clarg': '--confidence-lvl',
                'help': 'Confidence cutoff level as a floating point number',
                'type': float},
    'conftype': {'driverattr': 'conftype', 'clarg': '--confidence-better',
                 'help': 'Confidence type to define if higher or lower score '
                 'is better. One of [higher, lower]', 'type': str,
                 'choices': ['higher', 'lower']},
    'matchfilename': {'driverattr': 'matchfilename', 'clarg': '--matchcolname', 'type': str,
        'help': 'Column header field name for the column where matching accessions will be in'
        },
    'minpurity': {'driverattr': 'min_purity', 'clarg': '--min-precursor-purity',
        'type': float, 'required': False, 'default': 0,
        'help': 'Minimum purity of precursor required to output isobaric '
        'quantification data, MS2 scans with purity below this value '
        'will be assigned NA in isobaric channels'},
    'medianpsms': {'driverattr': 'medianpsms', 'clarg': '--medianpsms',
                   'help': 'In case of using a separate PSM table with more '
                   'data to generate more robust medians (i.e. a superset of '
                   'input PSMs), specify that file here. The normalization '
                   'factors will be calculated on this file, and the PSMs '
                   'in the input will be adjusted using those factors rather '
                   'than factors derived from solely their own quantification '
                   'data', 'type': 'file', 'required': False},
    'percofn': {'driverattr': 'percofn', 'clarg': '--perco', 'help': 'Percolator '
        'XML output file', 'type': 'file'},
    'qvalitypsms': {'driverattr': 'qvalitypsms', 'clarg': '--qvalitypsms', 'help': 'Qvality '
        'tab separated output containing targets and decoys (i.e. prepared with qvality -d) '
        'for PSM scores', 'type': 'file', 'required': False},
    'qvalitypeps': {'driverattr': 'qvalitypeps', 'clarg': '--qvalitypeps', 'help': 'Qvality '
        'tab separated output containing targets and decoys (i.e. prepared with qvality -d) '
        'for peptide scores ', 'type': 'file', 'required': False},
    'mzidfns': {'driverattr': 'mzidfns', 'clarg': '--mzids', 'help': 'MzIdentML '
        ' output files belonging to PSM table TSV files, use same order as for TSVs', 
        'type': 'file', 'nargs': '+'},
    'peptidecolpattern': {'driverattr': 'peptidecolpattern',
        'type': str, 'clarg': '--peptidecolpattern', 'default': False,
        'required': False, 'help': 'Regular expression pattern to find header '
        'field in table where peptide sequences are stored'},
    'splitcol': {'driverattr': 'splitcol', 'clarg': '--splitcol',
                 'help': 'Either a column number to split a PSM table on, or '
                 '"TD", "bioset" for splitting on target/decoy or biological '
                 'sample set columns (resulting from msstitch perco2psm or '
                 'msstitch psmtable. First column is number 1.'
                 },
    'lucitemplate': {'driverattr': 'lucitemplate', 'clarg': '--lucitemplate', 'type': 'file',
            'help': 'Template for luciphor config file if not using the default one found '
            'in .. ', # FIXME get default luci template here',
            },
    'mods': {'driverattr': 'mods', 'clarg': '--mods', 'nargs': '+', 'type': str,
            'help': 'All modification names used in the experiment, e.g. Phospho, '
            'Oxidation, TMT10plex'
            },
    'luciptms': {'driverattr': 'luciptms', 'clarg': '--luciptms', 'nargs': '+', 'type': str,
            'help': 'PTM names (unimod) and amino acid residues to run luciphor against on '
            'the PSM table of a sample(set). Specified as either --luciptms Phospho, or as '
            '--luciptms Phospho:ST . In the first case the default amino acids will be '
            'used (here STY), which can be found in ', ## FIXME get defaults
            },
    'ms2tolerance': {'driverattr': 'ms2tol', 'clarg': '--ms2tol', 'type': str,
            'help': 'MS2 fragment ion tolerance for Luciphor to use, default is 0.025Da, '
            'valid units are Da and ppm (e.g. --ms2tol 10ppm).', 'default': '0.025Da',
            },
    'luci_minpsms': {'driverattr': 'luci_minpsms', 'clarg': '--luciminpsms', 'type': int,
            'help': 'Minimum amount of PSMs needed for any charge state in order to build '
            'a model for that state in Luciphor',
            },
    'luci_frag': {'driverattr': 'luci_frag', 'clarg': '--fragmentation', 'type': str,
        'choices': ['hcd', 'cid'], 'default': 'hcd', 'help': 'Fragmentation type, one of HCD, CID',
        },
    'maxlength': {'driverattr': 'maxlength', 'default': 50,
                  'help': 'Maximum length of peptide to be included in Luciphor analysis',
                  'type': int, 'clarg': '--maxlen', 'required': False},
    'maxcharge': {'driverattr': 'maxcharge', 'default': 6,
                  'help': 'Maximum charge of PSM to be included in Luciphor analysis',
                  'type': int, 'clarg': '--maxcharge', 'required': False},
    'lucimemory': {'driverattr': 'lucimem', 'type': int, 'clarg': '--luciXmx',
            'help': 'Java heap memory maximum for luciphor (in GB, default 1)', 'default': 1,
            },
    'lucithread': {'driverattr': 'lucithread', 'type': int, 'clarg': '--lucithread',
            'help': 'Amount of threads to run Luciphor on, default 1', 'default': 1,
            },
    'mzmlpath': {'driverattr': 'mzmlpath', 'type': str, 'clarg': '--mzmlpath', 'default': '.',
            'help': 'Path to where mzML files can be found',
            },
}
psmtable_options['quantcolpattern'] = {
    k: v for k, v in shared_options['quantcolpattern'].items()}
psmtable_options['quantcolpattern']['required'] = True
psmtable_options['featcol'] = {k: v for k, v 
        in shared_options['featcol'].items()}
psmtable_options['featcol'].update(
        {'help': 'Column number in table in which desired '
        'accessions to summarize are stored. First column number is 1. If not '
        'specified this will summarize to PSMs themselves, i.e. only calculate '
        'the ratios and add those to the PSM table.'})

pepprottable_options = {
    # a mock infile to make sure we don't show or need an infile, e.g. in
    # case of building something from lookup
    'mock_infn': {'driverattr': 'fn', 'clarg': '-i', 'required': False,
                  'help': argparse.SUPPRESS, 'type': 'file'},
    'fdr': {'driverattr': 'fdr', 'clarg': '--fdr', 'action': 'store_const',
            'default': False, 'const': True, 'required': False,
            'help': 'Output FDR data to table'},
    'quantfile': {'driverattr': 'quantfile', 'clarg': '--quantfile',
                  'type': 'file', 'help': 'File containing isobaric quant '
                  'data to add to table.'},
    'scorecolpattern': {'driverattr': 'scorecolpattern', 'type': str,
                        'clarg': '--scorecolpattern', 'help': 'Regular '
                        'expression pattern to find column where score '
                        'to use (e.g. percolator svm-score) is written.'},

}


prottable_options = {k: v for k, v in pepprottable_options.items()}
prottable_options.update({
    'psmfile': {'driverattr': 'psmfile', 'clarg': '--psmtable', 'type': 'file',
                'help': 'PSM table file containing isobaric quant data to '
                'add to table.', 'conditional_required': ['quantcolpattern']},
    'decoyfn': {'driverattr': 'decoyfn',
                'help': 'Decoy peptide table input file',
                'type': 'file', 'clarg': '--decoyfn'},
    'minlogscore': {'driverattr': 'minlogscore', 'clarg': '--logscore',
                    'action': 'store_const', 'default': False, 'const': True,
                    'required': False, 'help': 'Score, e.g. q-values will '
                    'be converted to -log10 values.'},
    't_fasta': {'driverattr': 't_fasta', 'clarg': '--targetfasta', 'required': False,
                'type': 'file', 'help': 'FASTA file with target proteins '
                'to determine best scoring proteins of target/decoy pairs '
                'for picked FDR. Required when using --fdrtype picked',},
    'd_fasta': {'driverattr': 'd_fasta', 'clarg': '--decoyfasta', 'required': False,
                'type': 'file', 'help': 'FASTA file with decoy proteins '
                'to determine best scoring proteins of target/decoy pairs '
                'for picked FDR. Required when using --fdrtype picked',},
    'fdrtype': {'driverattr': 'fdrtype', 'clarg': '--fdrtype', 'required': False,
                'type': str, 'choices': ['picked', 'classic'], 'default': 'classic',
                 'help': 'FDR strategy type used. Can be one of [classic, picked]. '
                 'Picked FDR is implemented after Savitski et al. 2015, MCP, and '
                 'needs target and decoy fasta files to form pairs',
                 },
    'mergecutoff': {'driverattr': 'mergecutoff', 'clarg': '--mergecutoff',
                    'type': float, 'default': False, 'help': 'FDR cutoff when '
                    'building merged protein table, to use when a cutoff has '
                    'been used before storing the table to lookup. FDR values '
                    'need to be stored in the lookup', 'required': False},
})


peptable_options = {k: v for k, v in pepprottable_options.items()}
peptable_options.update({
    'spectracol': {'driverattr': 'spectracol',
                   'type': int, 'clarg': '--spectracol', 'help':
                   'Specify this column number (first col. is 1) '
                   'containing PSM table spectrafiles (e.g. mzML) '
                   'if you want to track PSMs when creating peptide '
                   'tables', 'required': False},
    'genecentric': {'driverattr': 'genecentric',
                    'clarg': '--genecentric', 'action': 'store_const',
                    'const': True, 'default': False, 'required': False,
                    'help': 'For peptide table merging. Do not include protein group '
                    'data in output, but use gene names instead to count peptides '
                    'per feature, determine peptide-uniqueness.',
                    },
    'modelqvals': {'driverattr': 'modelqvals',
                    'clarg': '--modelqvals', 'action': 'store_const',
                    'const': True, 'default': False, 'required': False,
                    'help': 'Create linear-modeled q-vals for peptides, to avoid '
                    'overlapping stepped low-qvalue data of peptides with '
                    'different scores', },
    'qvalthreshold': {'driverattr': 'qvalthreshold',
        'type': float, 'clarg': '--qvalthreshold', 'help': 'Specifies the '
        'inclusion threshold for q-values to fit a linear model to. Any scores/'
        'q-values below this threshold will not be used.', 'default': 10e-4,
        'required': False},
    'pepcolpattern': {'driverattr': 'pepcolpattern', 'required': False,
                      'clarg': '--pepcolpattern', 'type': str,
                      'default': None, 'help': 'Unique text pattern to identify '
                      'PEP column (given by percolator for peptides) in input table.'},
    'minpeptidenr': {'driverattr': 'minpeptidenr',
        'type': int, 'clarg': '--minpepnr', 'default': 10, 'help': 'Specifies '
        'the minimal amount of peptides (passing the --qvalthreshold) needed '
        'to fit a linear model, default is 10.', 'required': False},
    'totalprotfn': {'driverattr': 'totalprotfn', 'type': 'file', 
        'clarg': '--totalproteome', 'required': False,
        'help': 'File containing total proteome quantification to normalize PTM '
        'peptide quantification against, i.e. Phospho peptides isobaric quant '
        'ratios are divided by their proteins to distinguish differential '
        'phosphorylation from the protein expression variation in the sample. '
        'This file can also be a gene names or ENSG table. Accession should be '
        'in the first column. The file is preferably generated from a search '
        'without the relevant PTM, and can be a median-center normalized table. '
        },
})



from statistics import median
from numpy import polyfit
from math import log
import sys

from app.dataformats import mzidtsv as mzidtsvdata
from app.dataformats import peptable as peptabledata
from app.readers import tsv as reader


def get_quantcols(pattern, oldheader, coltype):
    """Searches for quantification columns using pattern and header list.
    Calls reader function to do regexp. Returns a single column for
    precursor quant."""
    if pattern is None:
       return False
    if coltype == 'precur':
        return reader.get_cols_in_file(pattern, oldheader, single_col=True)


def generate_peptides(tsvfn, oldheader, switch_map, scorecol, precurquantcol,
        fncol=None, higherbetter=True):
    if fncol is None:
        fncol = mzidtsvdata.HEADER_SPECFILE
    peptides = {}
    for psm in reader.generate_split_tsv_lines(tsvfn, oldheader):
        for oldkey, newkey in switch_map.items():
            try:
                psm[newkey] = psm.pop(oldkey)
            except KeyError:
                pass
        pepseq = psm[peptabledata.HEADER_PEPTIDE]
        peptides = evaluate_peptide(peptides, psm, pepseq, higherbetter,
                                    scorecol, fncol)
        add_quant_values(peptides, psm, precurquantcol)
    for peptide in peptides.values():
        peptide['line'][peptabledata.HEADER_LINKED_PSMS] = '; '.join(
            peptide['psms'])
        for qtype, pepquant in peptide['quant'].items():
            peptide['line'].update(parse_quant_data(qtype, pepquant))
        yield peptide['line']


def parse_quant_data(qtype, pepquant):
    if qtype == 'precur':
        quants = {peptabledata.HEADER_AREA: get_peptide_quant(pepquant, 'precur')}
    return quants


def get_peptide_quant(quantdata, quanttype):
    """Parses lists of quantdata and returns maxvalue from them. Strips NA"""
    parsefnx = {'precur': max}
    quantfloats = []
    for q in quantdata:
        try:
            quantfloats.append(float(q))
        except(TypeError, ValueError):
            pass
    if not quantfloats:
        return 'NA'
    return str(parsefnx[quanttype](quantfloats))


def add_quant_values(allpeps, psm, precurq_field):
    try:
        quants = allpeps[psm[peptabledata.HEADER_PEPTIDE]]['quant']
    except KeyError:
        quants = {}
    if precurq_field:
        try:
            quants['precur'].append(psm[precurq_field])
        except KeyError:
            quants['precur'] = [psm[precurq_field]]
    try:
        allpeps[psm[peptabledata.HEADER_PEPTIDE]]['quant'] = quants
    except KeyError:
        # PSM sequence does not exist in allpeps because has no good-scoring peptides
        pass


def recalculate_qvals_linear_model(peptides, scorecol, qvalthreshold, minpeptidenr):
    peptides = [x for x in peptides]
    slope, intercept = fit_linear_model(peptides, scorecol, qvalthreshold,
            minpeptidenr)
    for peptide in peptides:
        outpeptide = {k: v for k, v in peptide.items()}
        score = float(outpeptide[scorecol])
        if not slope:
            qval = 'NA'
        else:
            qval = 10 ** (slope * score + intercept)
        outpeptide[peptabledata.HEADER_QVAL_MODELED] = str(qval)
        yield outpeptide


def fit_linear_model(peptides, scorecol, qvalthreshold, minpeptidenr):
    pepq = []
    pepscore = []
    for peptide in peptides:
        try:
            qval = float(peptide[peptabledata.HEADER_QVAL])
        except KeyError:
            print('PSM table has no {} column, cannot perform linear modeling '
                    'of q-values. Exiting.'.format(peptabledata.HEADER_QVAL))
            sys.exit(1)
        except ValueError:
            print('Q-value found which is not a number: {}, please check the '
                    'PSM table. Exiting.'.format(peptide[peptabledata.HEADER_QVAL]))
            sys.exit(1)
        if qval > qvalthreshold:
            pepq.append(log(qval, 10))
            pepscore.append(float(peptide[scorecol]))
    if len(pepq) < minpeptidenr:
        slope, intercept = False, False
        print('Could not fit linear model through q-values, as only {} q-values '
                '({} are needed) was/were above the q-value threshold of {} '
                'for modeling inclusion (model is created from highest '
                'q-values to approximate the stepped lower ones)'.format(
                    len(pepq), minpeptidenr, qvalthreshold))
    else:
        slope, intercept = polyfit(pepscore, pepq, deg=1)
        print('Fitted linear model through qvalues (above {}) vs score. '
              'Slope={}, intercept={}'.format(qvalthreshold, slope, intercept))
    return slope, intercept


def add_peptide(allpeps, psm, key, score, fncol=False, new=False):
    peptide = {'score': score,
               'line': psm,
               'psms': []
               }
    if fncol:
        if not new:
            peptide['psms'] = allpeps[key]['psms']
    allpeps[key] = peptide


def add_psm_link(peptide, psm, fncol):
    peptide['psms'].append('{0}_{1}'.format(psm[fncol],
                                            psm[mzidtsvdata.HEADER_SPECSCANID]))


def evaluate_peptide(peptides, psm, key, higherbetter, scorecol, fncol=False):
    try:
        score = float(psm[scorecol])
    except ValueError:
        # If score is NA or similar, dont use this PSM
        return peptides
    try:
        existing_score = peptides[key]['score']
    except KeyError:
        add_peptide(peptides, psm, key, score, fncol, new=True)
    else:
        if higherbetter and score > existing_score:
            add_peptide(peptides, psm, key, score, fncol)
        elif not higherbetter and score < existing_score:
            add_peptide(peptides, psm, key, score, fncol)
    if fncol:
        add_psm_link(peptides[key], psm, fncol)
    return peptides



from app.drivers import base
from app.readers import tsv as tsvreader
from app.actions import merge

from app.drivers.options import prottable_options, peptable_options, lookup_options
from app.dataformats import prottable as ph 
from app.dataformats import peptable as peph

class MergeDriver(base.PepProttableDriver):
    outsuffix = ''
    command = 'merge'
    commandhelp = ('Create merged gene/protein/peptide tables from multiple '
            'sample set tables')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['setnames',
            'quantcolpattern', 'precursorquantcolpattern', 'fdrcolpattern',
            'flrcolpattern', 'multifiles', 'nogroup', 
            'featcol'], lookup_options))
        self.options.update(self.define_options(['lookupfn', 'mergecutoff'],
            prottable_options))
        self.options.update(self.define_options(['pepcolpattern',
            'genecentric'], peptable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        header = tsvreader.get_tsv_header(self.fn[0])
        self.header = [header[0]]
        if header[0] != peph.HEADER_PEPTIDE:
            self.is_peptidetable = False
            self.pepcolpattern = None # override input if any
        #else:
        if header[0] == peph.HEADER_PEPTIDE:
            self.header.append(peph.HEADER_BAREPEP)
            self.is_peptidetable = True
            if self.genecentric:
                self.lookuptype = 'peptidegenecentrictable'
                self.header.extend([peph.HEADER_GENES, peph.HEADER_ASSOCIATED])
            elif self.nogroup:
                self.lookuptype = 'peptidetableplain'
                self.header.extend([peph.HEADER_PROTEINS, peph.HEADER_STARTSTOP])
            else:
                self.header.extend([
                    peph.HEADER_PROTEINS, peph.HEADER_STARTSTOP, peph.HEADER_NO_CONTENTPROTEINS, 
                    peph.HEADER_DESCRIPTIONS, peph.HEADER_COVERAGES, peph.HEADER_GENES, 
                    peph.HEADER_ASSOCIATED])
                self.lookuptype = 'peptidetable'
        elif header[0] == ph.HEADER_PROTEIN and self.nogroup:
            self.lookuptype = 'plainprottable'
        elif header[0] == ph.HEADER_PROTEIN:
            self.lookuptype = 'prottable'
            self.header.extend([
                ph.HEADER_GENEID, ph.HEADER_GENENAME, ph.HEADER_DESCRIPTION,
                ph.HEADER_COVERAGE, ph.HEADER_CONTENTPROT, ph.HEADER_NO_PROTEIN])
        elif header[0] == ph.HEADER_GENENAME:
            self.lookuptype = 'associdtable'
            self.header.extend([
                ph.HEADER_GENEID, ph.HEADER_PROTEINS, ph.HEADER_DESCRIPTION])
        elif header[0] == ph.HEADER_GENEID:
            self.lookuptype = 'genetable'
            self.header.extend([
                ph.HEADER_GENENAME, ph.HEADER_PROTEINS, ph.HEADER_DESCRIPTION])

    def set_features(self):
        # FIXME need check to see same poolnames correlate with self.fn len
        self.initialize_lookup()
        self.featcol = self.featcol - 1
        self.setnames = [x.replace('"', '') for x in self.setnames]
        merge.create_lookup(self.fn, self.lookup, self.setnames, self.featcol,
                self.precursorquantcolpattern, self.quantcolpattern,
                self.fdrcolpattern, self.pepcolpattern, self.flrcolpattern)
        for field in self.lookup.stdheaderfields:
            self.header.extend(['{}_{}'.format(x, field) for x in self.setnames])
        if self.fdrcolpattern and not self.is_peptidetable:
            self.header.extend(['{}_{}'.format(x, ph.HEADER_QVAL) for x in self.setnames])
        if self.is_peptidetable:
            if self.flrcolpattern:
                self.header.extend(['{}_{}'.format(x, peph.HEADER_FALSE_LOC_RATE) for x in self.setnames])
            if self.precursorquantcolpattern:
                self.header.extend(['{}_{}'.format(x, peph.HEADER_AREA) for x in self.setnames])
        elif self.precursorquantcolpattern:
            self.header.extend(['{}_{}'.format(x, ph.HEADER_AREA) for x in self.setnames])
        if self.quantcolpattern:
            stored_psmnrs = self.lookup.check_isoquant_psmnrs()
            channels = [x for x in self.lookup.get_isoquant_headernames(stored_psmnrs)]
            self.header.extend([x[0] for x in channels])
            if stored_psmnrs:
                self.header.extend([f'{setn}_{ph.HEADER_NO_FULLQ_PSMS}' for setn in self.setnames])
                self.header.extend([x[1] for x in channels])
        self.features = merge.build_proteintable(self.lookup, self.mergecutoff, self.quantcolpattern)



import sys

from app.readers import tsv as tsvreader
from app.dataformats import mzidtsv as mzidtsvdata


def merge_mzidtsvs(fns, header):
    for fn in fns:
        if header != tsvreader.get_tsv_header(fn):
            raise RuntimeError('Headers of TSV files to concatenate are '
                               'not identical')
    for psm in tsvreader.generate_tsv_lines_multifile(fns, header):
        yield psm


def get_splitfield(header, splitcol):
    """Returns column nr on which to split PSM table. Chooses from flags
    given via bioset and splitcol"""
    if splitcol == 'bioset':
        return mzidtsvdata.HEADER_SETNAME
    elif splitcol == 'TD':
        return mzidtsvdata.HEADER_TARGETDECOY
    else:
        try:
            splitcol = int(splitcol)
        except ValueError:
            print('ERROR: --splitcol must be an integer or "TD", or "bioset"')
            sys.exit(1)
        else:
            return header[splitcol - 1]


def generate_psms_split(psms, splitfield):
    """Loops PSMs and outputs dictionaries passed to writer. Dictionaries
    contain the PSMs and info to which split pool the
    respective PSM belongs"""
    for psm in psms:
        yield {'psm': psm,
               'split_pool': psm[splitfield],
               }



from Bio import SeqIO
from collections import defaultdict

from app.actions.sequence import trypsinize

PROTEIN_STORE_CHUNK_SIZE = 100000


def create_searchspace_wholeproteins(lookup, fastafn, minpeplen):
    fasta = SeqIO.parse(fastafn, 'fasta')
    storeseqs = {}
    peptotal, pepsubtotal = 0, 0
    for record in fasta:
        prot_id, protseq = record.id, str(record.seq)
        storeseqs[prot_id] = {'seq': protseq, 'peps': []}
        isoseq = protseq.replace('L', 'I') 
        for pos in range(0, len(protseq) - minpeplen + 1):
            possible_pep = isoseq[pos:pos + minpeplen]
            pepsubtotal += 1
            storeseqs[prot_id]['peps'].append((possible_pep, pos))
        if pepsubtotal > PROTEIN_STORE_CHUNK_SIZE:
            lookup.store_pep_proteins(storeseqs)
            peptotal += pepsubtotal
            pepsubtotal = 0
            storeseqs = {}
    peptotal += pepsubtotal
    lookup.store_pep_proteins(storeseqs)
    lookup.index_proteins()
    print(f'Stored {peptotal} peptides')


def process_sequence(seq, do_trypsinize, proline_cut, miss_cleavage, minlen, reverse, ntermmloss): 
    if do_trypsinize:
        pepseqs = trypsinize(seq, proline_cut, miss_cleavage=miss_cleavage, split_stop_codons=True,
                opt_nt_meth_loss=ntermmloss)
    else:
        pepseqs = [(seq)]
    # Exchange all leucines to isoleucines because MS can't differ
    pepseqs = (pep.replace('L', 'I') for pep in pepseqs)
    if minlen:
        pepseqs = (pep for pep in pepseqs if len(pep) >= minlen)
    if reverse:
        pepseqs = (pep[::-1] for pep in pepseqs)
    return pepseqs


def create_searchspace(lookup, infile, minlen, proline_cut=False, reverse_seqs=True,
        do_trypsinize=True, miss_cleavage=False, ntermmloss=False):
    """Given a FASTA database, proteins are trypsinized and resulting peptides
    stored in a database or dict for lookups"""
    with open(infile) as fp:
        ftype = 'fasta' if fp.read(1) == '>' else 'txt'
        fp.seek(0)
        if ftype == 'fasta':
            records = (str(x.seq) for x in SeqIO.parse(fp, 'fasta'))
        elif ftype == 'txt':
            records = (x.strip('\n') for x in fp)
        lookup.write_peps((pep,) for x in records for pep in process_sequence(x, do_trypsinize,
            proline_cut, miss_cleavage, minlen, reverse_seqs, ntermmloss))
    lookup.index_peps(reverse_seqs)


def create_searchspace_map_accessions(lookup, infile, minlen, proline_cut=False, reverse_seqs=True,
        do_trypsinize=True, miss_cleavage=False, ntermmloss=False):
    """Given a FASTA database, proteins are trypsinized and resulting peptides
    stored in a database or dict for lookups, but this time with mapping protein accessions
    also stored, so we can find them later"""
    allpeps = defaultdict(list)
    peptotal, pepsubtotal = 0, 0
    with open(infile) as fp:
        for rec in SeqIO.parse(fp, 'fasta'):
            peps = process_sequence(rec.seq, do_trypsinize, proline_cut, miss_cleavage, minlen,
                reverse_seqs, ntermmloss)
            allpeps[rec.id].extend(peps)
    lookup.store_tryp_peps_mapped(allpeps) 
    lookup.index_peps_mapped(reverse_seqs)



import os
import sys

from app.lookups import base as lookups
from app.drivers.options import (shared_options, lookup_options, psmtable_options)

from app.readers import tsv as tsvreader
from app.readers import percolator as percoreaders
from app.readers import xml

from app.writers import tsv as tsvwriter
from app.writers import percolator as percowriters


class BaseDriver(object):
    tabletypes = []

    def __init__(self):
        self.lookupfn = None
        self.infiletype = ''

    def set_lookup(self):
        if self.lookupfn is not None and hasattr(self, 'lookuptype'):
            self.lookup = lookups.get_lookup(self.lookupfn, self.lookuptype)
        else:
            self.lookup = None

    def initialize_lookup(self, outfile=None):
        if self.lookup is None:
            # FIXME MUST be a set or mzml lookup? here is place to assert
            # correct lookuptype!
            if outfile is None and self.outfile is None:
                self.outfile = os.path.join(self.outdir,
                                            'mslookup_db.sqlite')
                lookupfile = self.outfile
            elif outfile is not None:
                lookupfile = outfile
            elif self.outfile is not None:
                lookupfile = self.outfile
            self.lookup = lookups.create_new_lookup(lookupfile,
                                                    self.lookuptype)
        self.lookup.add_tables(self.tabletypes)

    def set_options(self):
        self.options = self.define_options(['fn', 'outdir', 'outfile'], {})
        #self.options['-i']['help'] = self.options['-i']['help'].format(
        #@    self.infiletype)

    def get_commandhelp(self):
        return self.commandhelp

    def define_options(self, names, parser_options=None):
        """Given a list of option names, this returns a list of dicts
        defined in all_options and self.shared_options. These can then
        be used to populate the argparser with"""
        if parser_options is None:
            parser_options = {}
        options = {}
        for name in names:
            try:
                option = {k: v for k, v in parser_options[name].items()}
            except KeyError:
                option = {k: v for k, v in shared_options[name].items()}
            try:
                options.update({name: option})
                #options.update({option['clarg']: option})
            except TypeError:
                #options.update({option['clarg'][0]: option})
                options.update({name: option})
        return options

    def parse_input(self, **kwargs):
        # Set option values
        for option in self.options.values():
            opt_argkey = option['driverattr']
            opt_val = kwargs.get(opt_argkey)
            setattr(self, opt_argkey, opt_val)
        # Parse conditional required options
        required = {}
        for optname, option in self.options.items():
            key = option['driverattr']
            conds = option.get('conditional_required', [])
            for cond in conds:
                if getattr(self, self.options[cond]['driverattr']) and not getattr(self, key):
                    try:
                        required[cond].append(optname) 
                    except KeyError:
                        required[cond] = [optname]
        for cond, keys in required.items():
            print('When using {}, you must also specify {}'.format(
                self.options[cond]['clarg'], ', '.join([self.options[k]['clarg'] for k in keys])))
            sys.exit(1)
        # Prepare for output
        if self.outdir is None:
            self.outdir = os.getcwd()
        if self.outfile is not None:
            if not os.path.isabs(self.outfile):
                self.outfile = os.path.join(self.outdir, self.outfile)

    def start(self, **kwargs):
        self.parse_input(**kwargs)
        self.set_lookup()
        self.run()

    def create_outfilepath(self, fn, suffix=None):
        if self.outfile is None:
            basefn = os.path.basename(fn)
            outfn = os.path.join(self.outdir, basefn + suffix)
        else:
            outfn = self.outfile
        return outfn

    def number_to_headerfield(self, columnr, header):
        return header[int(columnr) - 1]

    def get_column_header_for_number(self, column_var_names, header=False):
        """This function subtracts 1 from inputted column number to comply
        with programmers counting (i.e. from 0, not from 1). For TSV data."""
        if not header:
            header = self.oldheader
        for col in column_var_names:
            value = getattr(self, col)
            if not value or value is None:
                continue
            setattr(self, col, self.number_to_headerfield(value, header))

    def get_commandname(self):
        return self.command

    def get_fastadelim_genefield(self, delimtype, genefield):
        if delimtype is not None:
            fastadelim = {'tab': '\t',
                          'semicolon': ';', 'pipe': '|'}[delimtype]
        else:
            fastadelim = None
        if genefield is not None:
            genefield -= 1
        return fastadelim, genefield

    def prepare(self):
        pass

    def run(self):
        self.prepare()
        self.set_features()
        self.write()


class LookupDriver(BaseDriver):
    outfile, outdir = None, None
    tabletypes = []

    def __init__(self):
        super().__init__()
        self.parser_options = lookup_options

    def set_options(self):
        super().set_options()
        del(self.options['fn'])
        del(self.options['outfile'])
        del(self.options['outdir'])
        self.options.update(self.define_options(['lookupfn'],
                                                lookup_options))

    def run(self):
        self.initialize_lookup()
        self.create_lookup()


class PSMDriver(BaseDriver):
    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'
        self.parser_options = psmtable_options

    def prepare(self):
        if type(self.fn) == list:
            self.first_infile = self.fn[0]
        else:
            self.first_infile = self.fn
        self.oldheader = tsvreader.get_tsv_header(self.first_infile)
        self.oldpsms = tsvreader.generate_split_tsv_lines(self.fn, self.oldheader)

    def write(self):
        outfn = self.create_outfilepath(self.first_infile, self.outsuffix)
        tsvwriter.write_tsv(self.header, self.psms, outfn)


class PercolatorDriver(BaseDriver):
    """Driver for percolator functions"""
    def __init__(self):
        super().__init__()
        self.infiletype = 'percolator out XML'

    def prepare_percolator_output(self, fn):
        """Returns namespace and static xml from percolator output file"""
        ns = xml.get_namespace(fn)
        static = percoreaders.get_percolator_static_xml(fn, ns)
        return ns, static

    def get_all_peptides(self):
        return percoreaders.generate_peptides(self.fn, self.ns)

    def get_all_psms(self):
        return percoreaders.generate_psms(self.fn, self.ns)

    def get_all_psms_strings(self):
        return percoreaders.generate_psms_multiple_fractions_strings([self.fn],
                                                                self.ns)

    def get_all_peptides_strings(self):
        return percoreaders.generate_peptides_multiple_fractions_strings([self.fn],
                                                                    self.ns)

    def prepare(self):
        self.ns, self.static_xml = self.prepare_percolator_output(self.fn)
        self.allpeps = self.get_all_peptides()
        self.allpsms = self.get_all_psms()

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        percowriters.write_percolator_xml(self.static_xml, self.features, outfn)


class PepProttableDriver(BaseDriver):

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        tsvwriter.write_table_with_na(self.header, self.features, outfn)



from Bio import SeqIO


def get_proteins_for_db(fastafn, fastadelim, genefield):
    """Runs through fasta file and returns proteins accession nrs, sequences
    and evidence levels for storage in lookup DB. Duplicate accessions in
    fasta are accepted and removed by keeping only the last one.
    """
    records = {acc: (rec, get_record_type(rec)) for acc, rec in 
            SeqIO.index(fastafn, 'fasta').items()}
    proteins = ((x,) for x in records.keys())
    sequences = ((acc, str(rec.seq)) for acc, (rec, rtype) in records.items())
    desc = ((acc, get_description(rec, rtype)) for acc, (rec, rtype) in records.items())
    evid = ((acc, get_uniprot_evidence_level(rec, rtype)) for acc, (rec, rtype) in 
        records.items())
    ensgs = [(get_ensg(rec), acc) for acc, (rec, rtype) in records.items()
            if rtype == 'ensembl']
    def sym_out():
        symbols = ((get_symbol(rec, rtype, fastadelim, genefield), acc) for 
                acc, (rec, rtype) in records.items() if rtype)
        othergene = ((get_other_gene(rec, fastadelim, genefield), acc) for acc, (rec, rtype) in records.items()
                if not rtype and fastadelim and fastadelim in rec.description)
        yield from symbols
        yield from othergene
    return proteins, sequences, desc, evid, ensgs, [x for x in sym_out()]


def parse_fasta(fn):
    with open(fn) as fp:
        for record in SeqIO.parse(fp, 'fasta'):
            yield record


def get_record_type(record):
    dmod = get_decoy_mod_string(record.id)
    test_name = record.id
    if dmod is not None:
        test_name = record.id.replace(dmod, '')
    if test_name.split('|')[0] in ['sp', 'tr']:
        return 'swiss'
    elif test_name[:3] == 'ENS':
        return 'ensembl'
    else:
        return False


def get_decoy_mod_string(protein):
    mods = ['tryp_reverse', 'reverse', 'decoy', 'random', 'shuffle']
    for mod in mods:
        if mod in protein:
            if protein.endswith('_{}'.format(mod)):
                return '_{}'.format(mod)
            elif protein.endswith('{}'.format(mod)):
                return mod
            elif protein.startswith('{}_'.format(mod)):
                return '{}_'.format(mod)
            elif protein.startswith('{}'.format(mod)):
                return mod


def get_description(record, rectype):
    if rectype == 'ensembl':
        desc_spl = [x.split(':') for x in record.description.split()]
        try:
            descix = [ix for ix, x in enumerate(desc_spl) if x[0] == 'description'][0]
        except IndexError:
            return 'NA'
        desc = ' '.join([':'.join(x) for x in desc_spl[descix:]])[12:]
        return desc
    elif rectype == 'swiss':
        desc = []
        for part in record.description.split()[1:]:
            if len(part.split('=')) > 1:
                break
            desc.append(part)
        return ' '.join(desc)
    elif record.description.strip() == record.id:
        return 'NA'
    else:
        # record is first part of definition until space, so strip it
        return record.description[len(record.id) + 1:].lstrip()


def get_other_gene(record, fastadelim, genefield):
    return record.description.split(fastadelim)[genefield]


def get_genes_pickfdr(fastafn, outputtype, fastadelim, genefield):
    """Called by protein FDR module for both ENSG and e.g. Uniprot"""
    for rec in parse_fasta(fastafn):
        rtype = get_record_type(rec)
        if outputtype == 'protein':
            yield rec.id
        elif rtype == 'ensembl' and outputtype == 'ensg':
            yield get_ensg(rec)
        elif outputtype == 'genename':
            yield get_symbol(rec, rtype, fastadelim, genefield)


def get_ensg(record):
    fields = [x.split(':') for x in record.description.split()]
    try:
        return [x[1] for x in fields if x[0] == 'gene' and len(x) == 2][0]
    except IndexError:
        raise RuntimeError('ENSEMBL detected but cannot find gene ENSG in fasta')


def get_symbol(record, rectype, fastadelim, genefield):
    if rectype == 'ensembl':
        fields = [x.split(':') for x in record.description.split()]
        sym = [x[1] for x in fields if x[0] == 'gene_symbol' and len(x) == 2]
    elif rectype == 'swiss':
        fields = [x.split('=') for x in record.description.split()]
        sym = [x[1] for x in fields if x[0] == 'GN' and len(x) == 2]
    elif fastadelim and fastadelim in record.description and genefield:
        return record.description.split(fastadelim)[genefield]
    else:
        return 'NA'
    try:
        return sym[0]
    except IndexError:
        return 'NA'


def get_uniprot_evidence_level(record, rtype):
    """Returns uniprot protein existence evidence level for a fasta header.
    Evidence levels are 1-5, but we return 5 - x since sorting still demands
    that higher is better."""
    if rtype != 'swiss':
        return -1
    for item in record.description.split():
        item = item.split('=')
        try:
            if item[0] == 'PE' and len(item) == 2:
                return 5 - int(item[1])
        except IndexError:
            continue
    return -1



import os
import sqlite3
from itertools import product

from tests.integration import basetests


class TestSplitProteinHeader(basetests.BaseTestPycolator):
    command = 'splitperco'
    suffix = ''

    def setUp(self):
        super().setUp()
        self.resultfn = None

    def test_splitprotein(self):
        protheaders = ['^ENSP000002', 'ENSP000002|^ENSP000003']
        self.suffix = ''
        options = ['--protheaders'] + protheaders
        self.run_command(options)
        ensp2_incl = os.path.join(self.workdir, self.infilename + '_h0.xml')
        ensp3_not_2 = os.path.join(self.workdir, self.infilename + '_h1.xml')
        first_expected = os.path.join(self.fixdir, 'perco.xml_h0.xml')
        second_expected = os.path.join(self.fixdir, 'perco.xml_h1.xml')
        incl_exp_contents = self.read_percolator_out(first_expected)
        excl_exp_contents = self.read_percolator_out(second_expected)
        incl_contents = self.read_percolator_out(ensp2_incl)
        excl_contents = self.read_percolator_out(ensp3_not_2)
        self.assertEqual(len(incl_contents['psms']),
                         len(incl_exp_contents['psms']))
        self.assertEqual(len(incl_contents['peptides']),
                         len(incl_exp_contents['peptides']))
        self.assertEqual(len(excl_contents['psms']),
                         len(excl_exp_contents['psms']))
        self.assertEqual(len(excl_contents['peptides']),
                         len(excl_exp_contents['peptides']))
        ns = self.get_xml_namespace(ensp2_incl)
        for feat in ['psms', 'peptides']:
            for el in incl_contents[feat]:
                self.assertIn(
                        'ENSP000002', [x.text[:10] for x in el.findall('{%s}protein_id' % ns['xmlns'])])
            for el in excl_contents[feat]:
                self.assertIn('ENSP000003', [x.text[:10] for x in el.findall('{%s}protein_id' % ns['xmlns'])])
                self.assertNotIn('ENSP000002', [x.text[:10] for x in el.findall('{%s}protein_id' % ns['xmlns'])])


class TestFilterKnown(basetests.BaseTestPycolator):
    command = 'filterperco'
    suffix = '_filtseq.xml'
    dbpath = 'seqs.db'

    def test_noflags(self):
        seqs = ['DSGTLQSQEAK']
        basetests.create_db(seqs)
        self.assert_seqs_correct(seqs)

    def test_ntermwildcards(self):
        seqs = ['DSGTLQSQEAK']
        falloff = 'LLLLLL'
        basetests.create_db([f'{falloff}{x}' for x in seqs], reverse=True)
        self.assert_seqs_correct(seqs, ['--insourcefrag', f'{len(falloff)+1}'])

    def test_deamidate(self):
        seqs = ['NSGTLQSQEAK']
        basetests.create_db(seqs)
        self.assert_seqs_correct(seqs, ['--deamidate'], deamidate=True)

    def test_fullprotein(self):
        seqs = ['SEEAHRAEQLQDAEEEK']
        proteins = [f'XXXXXXXXX{x}XXXXXXXXX' for x in seqs]
        basetests.create_db(proteins, fullprotein=True, minlen=6)
        self.assert_seqs_correct(seqs, ['--fullprotein', '--minlen', '6'], fullprot=True)
        
    def deamidate(self, sequence):
        '''Deamidation is N->D, so any sequences found with D can have originally been N
        and should be checked with N in DB as well'''
        aa_possible = [(aa,) if aa != 'D' else ('D', 'N') for aa in sequence]
        return list(''.join(aa) for aa in product(*aa_possible))

    def assert_seqs_correct(self, seqs_filter, flags=[], deamidate=False, fullprot=False):
        """Does the actual testing"""
        options = ['--dbfile', self.dbpath]
        options.extend(flags)
        self.run_command(options)
        result = self.get_psm_pep_ids_from_file(self.resultfn)
        origin = self.get_psm_pep_ids_from_file(self.infile[0])
        seq_in_ori = {'peptide_ids': False, 'psm_seqs': False}
        for feattype in ['peptide_ids', 'psm_seqs']:
            original_seqs = origin[feattype]
            result_seqs = result[feattype]
            db = sqlite3.connect(self.dbpath)
            for oriseq in original_seqs:
                thisseq_filter = False
                seq_dbcheck = self.strip_modifications(oriseq)
                if deamidate:
                    testseqs = self.deamidate(seq_dbcheck)
                else:
                    testseqs = [seq_dbcheck]
                for seq in testseqs:
                    if seq in seqs_filter:
                        thisseq_filter = True
                        seq_in_ori[feattype] = True
                        self.assertNotIn(oriseq, result_seqs)
                if not thisseq_filter:
                    self.assertIn(oriseq, result_seqs)
        self.assertTrue(all(seq_in_ori.values()))



class TestDedupPeptides(basetests.BaseTestPycolator):
    command = 'dedupperco'
    suffix = '_dedup.xml'

    def test_do_no_duplicate_peps(self):
        options = []
        self.run_command(options)
        perco_input = self.get_psm_pep_ids_from_file(self.infile[0])
        perco_output = self.get_psm_pep_ids_from_file(self.resultfn)
        self.assertEqual(perco_input['psm_ids'], perco_output['psm_ids'])
        self.assertEqual(perco_input['peptide_ids'], perco_output['peptide_ids'])
        
    def test_remove_duplicate(self):
        self.infile = [os.path.join(self.fixdir, 'perco_duplicate.xml')]
        options = []
        self.run_command(options)
        perco_input = self.get_psm_pep_ids_from_file(self.infile[0])
        perco_output = self.get_psm_pep_ids_from_file(self.resultfn)
        self.assertEqual(perco_input['psm_ids'], perco_output['psm_ids'])
        self.assertTrue(len(perco_input['peptide_ids']) > len(perco_output['peptide_ids']))
        self.assertEqual(set(perco_input['peptide_ids']), set(perco_output['peptide_ids']))

    def test_remove_duplicate_psms_too(self):
        self.infile = [os.path.join(self.fixdir, 'perco_duplicate.xml')]
        options = ['--includepsms']
        self.run_command(options)
        perco_input = self.get_psm_pep_ids_from_file(self.infile[0])
        perco_output = self.get_psm_pep_ids_from_file(self.resultfn)
        self.assertEqual(len(perco_input['psm_ids']) - 1, len(perco_output['psm_ids']))
        self.assertEqual(len(perco_input['psm_seqs']) - 1, len(perco_output['psm_seqs']))
        self.assertEqual(set(perco_input['psm_ids']), set(perco_output['psm_ids']))
        self.assertEqual(set(perco_input['psm_seqs']), set(perco_output['psm_seqs']))
        self.assertEqual(len(perco_input['peptide_ids']) - 1, len(perco_output['peptide_ids']))
        self.assertEqual(set(perco_input['peptide_ids']), set(perco_output['peptide_ids']))



from Bio import SeqIO
from Bio.Seq import Seq
from random import Random


def create_trypsinized(proteins, proline_cut, miss_cleavage, minlen, nterm_meth_loss,
        splitstopcodons):
    for proseq in proteins:
        trypnr = 0
        outpeps = trypsinize(str(proseq.seq), proline_cut, miss_cleavage, nterm_meth_loss,
                splitstopcodons)
        if minlen:
            outpeps = [pep for pep in outpeps if len(pep) >= minlen]
        for pep in outpeps:
            trypnr += 1
            trypseq = SeqIO.SeqRecord(Seq(pep), id='{}_{}'.format(proseq.id, trypnr),
                    name=proseq.name, description='')
            yield trypseq


def trypsinize(input_protein, proline_cut=False, miss_cleavage=0, opt_nt_meth_loss=False, split_stop_codons=False):
    # TODO add cysteine to non cut options
    """Trypsinize a protein sequence. Returns a list of peptides.
    Peptides include both cut and non-cut when P is behind a tryptic
    residue. Number of missed cleavages can be specified.
    When opt_nt_meth_loss is True, peptides from protein N-term will come with
    and without the protein N-term methionine residue.
    """
    outpeps, nterm_losspeps = [], []
    trypres = set(['K', 'R'])
    noncutters = set()
    if not proline_cut:
        noncutters.add('P')
    if split_stop_codons:
        protein_sequences = input_protein.split('*')
    else:
        protein_sequences = [input_protein]
    for proseq in protein_sequences:
        peptide, proseq_outpeps = '', []
        protein_has_ntermlosspep = False
        for i, aa in enumerate(proseq.upper()):
            peptide += aa
            try:
                next_aa = proseq[i + 1]
            except IndexError:
                next_aa = False
            if aa in trypres and next_aa not in noncutters:
                if not outpeps and opt_nt_meth_loss and peptide[0].upper() == 'M':
                    # Include protein N-term M loss peptide
                    nterm_losspeps.append(peptide[1:])
                    protein_has_ntermlosspep = True
                proseq_outpeps.append(peptide)  # do actual cut by storing peptides
                peptide = ''
        if len(peptide):
            proseq_outpeps.append(peptide)
            
        # Get missed cleavages by re-attaching tryptic peptides, extra round for
        # nterm_loss peptides
        ft_pep_amount = len(proseq_outpeps)
        for i in range(1, miss_cleavage + 1):
            full_tryp_outpeps = [x for x in proseq_outpeps]
            for j, pep in enumerate(full_tryp_outpeps):
                if j + i < ft_pep_amount:
                    proseq_outpeps.append(''.join([x for x in (proseq_outpeps[j:j + i + 1])]))
            if protein_has_ntermlosspep and i < ft_pep_amount:
                # Also get the nterm-loss/miscleav peps if used
                nterm_losspeps.append(f'{nterm_losspeps[0]}{"".join(full_tryp_outpeps[1:i+1])}')
        outpeps.extend(proseq_outpeps)
    return outpeps + nterm_losspeps


def tryp_rev(seq, lookup, do_trypsinize, miss_cleavage, minlen, max_shuffle, keep_target):
    if do_trypsinize:
        segments = trypsinize(seq, miss_cleavage=miss_cleavage)
    else:
        segments = [str(seq.seq)]
    final_seq = {}
    decoy_segs = {}
    for i, s in enumerate(segments):
        if len(s) > 1 :
            if s[-1] in ['R', 'K']:
                decoy_segs[i] = '{}{}'.format(s[:-1][::-1], s[-1])
            else:
                decoy_segs[i] = s[::-1]
        else:
            decoy_segs[i] = s
    nr_decoymatching = 0
    if minlen:
        decoy_segs = {k: (v if len(v) >=minlen else '') for k,v in decoy_segs.items() }
    nr_peptides = len([x for x in decoy_segs.values() if x != ''])
    random_seeded = Random(1) # with defined seed so it becomes reproducible
    if lookup is not None:
        shufflecount = 0
        targets, tests = True, {k: (v, 0, v) for k,v in decoy_segs.items()}
        while targets:
            targets = lookup.get_multi_seq([x[0] for x in tests.values()])
            for i, (s, shufcount, origdecoy) in [(k,v) for k,v in tests.items()]: # list comprehension to not have dict change during iteration
                if s not in targets:
                    decoy_segs[i] = tests.pop(i)[0]
                elif shufcount < max_shuffle:
                    nterm = list(s[:-1])
                    random_seeded.shuffle(nterm)
                    tests[i] = ('{}{}'.format(''.join(nterm), s[-1]), shufcount + 1, origdecoy)
                elif keep_target:
                    decoy_segs[i] = origdecoy
                    tests.pop(i)
                    nr_decoymatching += 1
                else:
                    tests.pop(i)
                    decoy_segs.pop(i)
                    nr_decoymatching += 1
    if decoy_segs and set(decoy_segs.values()) != {''}:
        seq.seq = Seq(''.join([decoy_segs[i] for i in range(0, max(decoy_segs)+1) if i in decoy_segs]))
        seq.id = 'decoy_{}'.format(seq.name)
        seq.description = 'decoy_{}'.format(seq.description)
    else:
        seq = False
    return seq, nr_decoymatching, nr_peptides


def prot_rev(seq):
    seq.id = 'decoy_{}'.format(seq.name)
    seq = seq[::-1]
    return seq


def create_decoy_fa(fastafn, method, lookup, trypsinize, miss_cleavage, minlen, max_shuffle,
        keep_target):
    outfasta = SeqIO.parse(fastafn, 'fasta')
    if method == 'prot_rev':
        for seq in outfasta:
            yield prot_rev(seq)
    if method == 'tryp_rev':
        decoymatching, nr_peptides = 0, 0
        outtryp = (tryp_rev(x, lookup, trypsinize, miss_cleavage, minlen, 
            max_shuffle, keep_target) for x in outfasta)
        for seq, nr_decoymatching, nr_pep in outtryp:
            decoymatching += nr_decoymatching
            nr_peptides += nr_pep
            if seq:
                yield seq
        if decoymatching:
            retain = '(retained those as tryptic reverse)' if keep_target else ''
            print(f'Unable to shuffle {decoymatching} decoys (of a total of {nr_peptides} decoy peptides) '
                    f'that matched target DB {retain}')



from app.drivers import base
from app.actions import sequence
from app.actions.lookups import sequence as seqlup
from app.drivers.options import lookup_options, sequence_options
from Bio import SeqIO


class DecoySeqDriver(base.BaseDriver):
    outsuffix = '_decoy.fa'
    lookuptype = 'searchspace'
    command = 'makedecoy'
    commandhelp = """Create a decoy database from a FASTA file.
    tryp_rev reverses tryptic peptides
    prot_rev reverses full protein
    Both make sure no decoys are accidentally identical to a target sequence,
    unless --ignore-target-hits is passed"""

    def __init__(self):
        super().__init__()
        self.infiletype = 'FASTA'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options([
            'fn', 'outfile', 'lookupfn', 'scramble', 'ignoretarget', 'trypsinize', 
            'miss_cleavage', 'minlength', 'max_shuffle', 'keep_target'], sequence_options))
        self.options['lookupfn'].update({'required': False, 'default': None})

    def run(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        if self.lookup is None and not self.ignoretarget:
            self.initialize_lookup('decoychecker.sqlite')
            seqlup.create_searchspace(self.lookup, self.fn, self.minlength, reverse_seqs=False,
                    miss_cleavage=self.miss_cleavage)
        decoyfa = sequence.create_decoy_fa(self.fn, self.scramble, self.lookup, self.trypsinize,
                self.miss_cleavage, self.minlength, self.max_shuffle, self.keep_target)
        with open(outfn, 'w') as fp:
            SeqIO.write(decoyfa, fp, 'fasta')


class TrypsinizeDriver(base.BaseDriver):
    outsuffix = '_tryp.fa'
    command = 'trypsinize'
    commandhelp = """Trypsinize a FASTA file"""

    def __init__(self):
        super().__init__()
        self.infiletype = 'FASTA'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options([
            'fn', 'outfile', 'miss_cleavage', 'minlength', 'proline',
            'ntermmloss', 'stopcodonsplit'], sequence_options))

    def run(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        with open(self.fn) as fp, open(outfn, 'w') as wfp:
            seqs = SeqIO.parse(fp, 'fasta')
            trypfa = sequence.create_trypsinized(seqs, self.proline, self.miss_cleavage, self.minlength,
                    self.ntermmloss, self.stopcodonsplit)
            SeqIO.write(trypfa, wfp, 'fasta')



from app.lookups.sqlite.base import ResultLookupInterface


class QuantDB(ResultLookupInterface):

    def add_tables(self, tabletypes):
        if 'isobaric' in tabletypes:
            self.create_tables(['isobaric_quant', 'isobaric_channels',
                'precursor_ion_fraction'])
        if 'ms1' in tabletypes:
            self.create_tables(['ms1_quant', 'ms1_align', 'ms1_fwhm'])

    def get_fnfeats(self, fn_id):
        cursor = self.get_cursor()
        return cursor.execute(
            'SELECT mz, feature_id, charge, retention_time '
            'FROM ms1_quant '
            'WHERE mzmlfile_id=? ORDER BY mz', (fn_id,))

    def store_channelmap(self, channels):
        '''Stores channels if they do not exist yet (unique names in table)'''
        self.store_many(
            'INSERT OR IGNORE INTO isobaric_channels(channel_name) VALUES(?)', channels)

    def store_isobaric_quants(self, quants, pifs):
        self.store_many(
            'INSERT INTO isobaric_quant(spectra_id, channel_id, intensity) '
            'VALUES (?, ?, ?)', quants)
        self.store_many(
                'INSERT INTO precursor_ion_fraction(spectra_id, pif) VALUES (?, ?)', pifs)

    def index_isobaric_quants(self):
        self.index_column('spectraid_index', 'isobaric_quant', 'spectra_id')
        self.index_column('channel_id_index', 'isobaric_quant', 'channel_id')
        self.index_column('pif_spectraid_ix', 'precursor_ion_fraction', 'spectra_id')

    def get_specmap(self, fn_id):
        """Returns all DB spectra ids for spectra filename, keyed by 
        mzML SID (per scan)"""
        cursor = self.get_cursor()
        values = [fn_id]
        sql = 'SELECT scan_sid,spectra_id FROM mzml WHERE mzmlfile_id=? '
        cursor.execute(sql, tuple(values))
        return {k: {'id': sid} for k, sid in cursor.fetchall()}

    def get_channelmap(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT channel_id, channel_name FROM isobaric_channels')
        return cursor

    def store_ms1_quants(self, quants):
        # Return feat IDs?
        quants = self.store_many_return_id(
            'INSERT INTO ms1_quant(mzmlfile_id, retention_time, mz, '
            'charge, intensity) VALUES (?, ?, ?, ?, ?)', quants)
        return quants

    def store_fwhm(self, quants):
        # FIXME get feature_id for all passed quants, possibly directly when storing them
        self.store_many('INSERT INTO ms1_fwhm(feature_id, fwhm) VALUES (?, ?)', quants)

    def store_ms1_alignments(self, aligns):
        self.store_many(
            'INSERT INTO ms1_align(spectra_id, feature_id) '
            'VALUES (?, ?)', aligns)

    def index_precursor_quants(self):
        self.index_column('ms1_feat_ix', 'ms1_quant', 'feature_id')
        self.index_column('charge_index', 'ms1_quant', 'charge')
        self.index_column('feat_mz_index', 'ms1_quant', 'mz')
        self.index_column('ms1_mzfn_index', 'ms1_quant', 'mzmlfile_id')
        self.index_column('fwhm_feat_index', 'ms1_fwhm', 'feature_id')

    def index_aligned_quants(self):
        self.index_column('ms1al_feat_ix', 'ms1_align', 'feature_id')
        self.index_column('ms1al_spec_ix', 'ms1_align', 'spectra_id')

    def get_spectra_mz_sorted(self):
        return self.get_cursor().execute(
            'SELECT spectra_id, mzmlfile_id, charge, mz, retention_time '
            'FROM mzml ORDER BY mzmlfile_id,mz')
